@@ -36,6 +36,7 @@ type Command struct {
 	BuildTime string
 
 	closing chan struct{}
+	pidfile string
 	Closed  chan struct{}
 
 	Stdin  io.Reader
@@ -44,6 +45,9 @@ type Command struct {
 	Logger zap.Logger
 
 	Server *Server
+
+	// How to get environment variables. Normally set to os.Getenv, except for tests.
+	Getenv func(string) string
 }
 
 // NewCommand return a new instance of Command.
@@ -67,7 +71,7 @@ func (cmd *Command) Run(args ...string) error {
 	}
 
 	// Print sweet InfluxDB logo.
-	fmt.Print(logo)
+	fmt.Fprint(cmd.Stdout, logo)
 
 	// Mark start-up in log.
 	cmd.Logger.Info(fmt.Sprintf("InfluxDB starting, version %s, branch %s, commit %s",
@@ -78,6 +82,7 @@ func (cmd *Command) Run(args ...string) error {
 	if err := cmd.writePIDFile(options.PIDFile); err != nil {
 		return fmt.Errorf("write pid file: %s", err)
 	}
+	cmd.pidfile = options.PIDFile
 
 	// Parse config
 	config, err := cmd.ParseConfig(options.GetConfigPath())
@@ -86,7 +91,7 @@ func (cmd *Command) Run(args ...string) error {
 	}
 
 	// Apply any environment variables on top of the parsed config
-	if err := config.ApplyEnvOverrides(); err != nil {
+	if err := config.ApplyEnvOverrides(cmd.Getenv); err != nil {
 		return fmt.Errorf("apply env config: %v", err)
 	}
 
@@ -96,8 +101,9 @@ func (cmd *Command) Run(args ...string) error {
 	}
 
 	if config.HTTPD.PprofEnabled {
-		// Turn on block profiling to debug stuck databases
+		// Turn on block and mutex profiling.
 		runtime.SetBlockProfileRate(int(1 * time.Second))
+		runtime.SetMutexProfileFraction(1) // Collect every sample
 	}
 
 	// Create server from config and start it.
@@ -128,6 +134,7 @@ func (cmd *Command) Run(args ...string) error {
 // Close shuts down the server.
 func (cmd *Command) Close() error {
 	defer close(cmd.Closed)
+	defer cmd.removePIDFile()
 	close(cmd.closing)
 	if cmd.Server != nil {
 		return cmd.Server.Close()
@@ -147,6 +154,14 @@ func (cmd *Command) monitorServerErrors() {
 	}
 }
 
+func (cmd *Command) removePIDFile() {
+	if cmd.pidfile != "" {
+		if err := os.Remove(cmd.pidfile); err != nil {
+			cmd.Logger.Error("unable to remove pidfile", zap.Error(err))
+		}
+	}
+}
+
 // ParseFlags parses the command line flags from args and returns an options set.
 func (cmd *Command) ParseFlags(args ...string) (Options, error) {
 	var options Options
@@ -172,8 +187,7 @@ func (cmd *Command) writePIDFile(path string) error {
 	}
 
 	// Ensure the required directory structure exists.
-	err := os.MkdirAll(filepath.Dir(path), 0777)
-	if err != nil {
+	if err := os.MkdirAll(filepath.Dir(path), 0777); err != nil {
 		return fmt.Errorf("mkdir: %s", err)
 	}
 
