@@ -40,6 +40,47 @@ func TestMarshal(t *testing.T) {
 	}
 }
 
+func TestMarshalFields(t *testing.T) {
+	for _, tt := range []struct {
+		name  string
+		value interface{}
+		exp   string
+	}{
+		{
+			name:  "Float",
+			value: float64(2),
+			exp:   `value=2`,
+		},
+		{
+			name:  "Integer",
+			value: int64(2),
+			exp:   `value=2i`,
+		},
+		{
+			name:  "Unsigned",
+			value: uint64(2),
+			exp:   `value=2u`,
+		},
+		{
+			name:  "String",
+			value: "foobar",
+			exp:   `value="foobar"`,
+		},
+		{
+			name:  "Boolean",
+			value: true,
+			exp:   `value=true`,
+		},
+	} {
+		t.Run(tt.name, func(t *testing.T) {
+			fields := map[string]interface{}{"value": tt.value}
+			if have, want := models.Fields(fields).MarshalBinary(), []byte(tt.exp); !bytes.Equal(have, want) {
+				t.Fatalf("unexpected field output: %s != %s", string(have), string(want))
+			}
+		})
+	}
+}
+
 func TestTags_HashKey(t *testing.T) {
 	tags = models.NewTags(map[string]string{"A FOO": "bar", "APPLE": "orange", "host": "serverA", "region": "uswest"})
 	got := tags.HashKey()
@@ -523,6 +564,7 @@ func TestParsePointBadNumber(t *testing.T) {
 		"cpu v=-e-e-e ",
 		"cpu v=42+3 ",
 		"cpu v= ",
+		"cpu v=-123u",
 	} {
 		_, err := models.ParsePointsString(tt)
 		if err == nil {
@@ -567,10 +609,17 @@ func TestParsePointMinInt64(t *testing.T) {
 	}
 
 	// min int
-	_, err = models.ParsePointsString(`cpu,host=serverA,region=us-west value=-9223372036854775808i`)
+	p, err := models.ParsePointsString(`cpu,host=serverA,region=us-west value=-9223372036854775808i`)
 	if err != nil {
 		t.Errorf(`ParsePoints("%s") mismatch. got %v, exp nil`, `cpu,host=serverA,region=us-west value=-9223372036854775808i`, err)
 	}
+	fields, err := p[0].Fields()
+	if err != nil {
+		t.Fatal(err)
+	}
+	if exp, got := int64(-9223372036854775808), fields["value"].(int64); exp != got {
+		t.Fatalf("ParsePoints Value mismatch. \nexp: %v\ngot: %v", exp, got)
+	}
 
 	// leading zeros
 	_, err = models.ParsePointsString(`cpu,host=serverA,region=us-west value=-0009223372036854775808i`)
@@ -587,10 +636,17 @@ func TestParsePointMaxFloat64(t *testing.T) {
 	}
 
 	// max float
-	_, err = models.ParsePointsString(fmt.Sprintf(`cpu,host=serverA,region=us-west value=%s`, string(maxFloat64)))
+	p, err := models.ParsePointsString(fmt.Sprintf(`cpu,host=serverA,region=us-west value=%s`, string(maxFloat64)))
 	if err != nil {
 		t.Errorf(`ParsePoints("%s") mismatch. got %v, exp nil`, `cpu,host=serverA,region=us-west value=9223372036854775807`, err)
 	}
+	fields, err := p[0].Fields()
+	if err != nil {
+		t.Fatal(err)
+	}
+	if exp, got := math.MaxFloat64, fields["value"].(float64); exp != got {
+		t.Fatalf("ParsePoints Value mismatch. \nexp: %v\ngot: %v", exp, got)
+	}
 
 	// leading zeros
 	_, err = models.ParsePointsString(fmt.Sprintf(`cpu,host=serverA,region=us-west value=%s`, "0000"+string(maxFloat64)))
@@ -607,10 +663,17 @@ func TestParsePointMinFloat64(t *testing.T) {
 	}
 
 	// min float
-	_, err = models.ParsePointsString(fmt.Sprintf(`cpu,host=serverA,region=us-west value=%s`, string(minFloat64)))
+	p, err := models.ParsePointsString(fmt.Sprintf(`cpu,host=serverA,region=us-west value=%s`, string(minFloat64)))
 	if err != nil {
 		t.Errorf(`ParsePoints("%s") mismatch. got %v, exp nil`, `cpu,host=serverA,region=us-west value=...`, err)
 	}
+	fields, err := p[0].Fields()
+	if err != nil {
+		t.Fatal(err)
+	}
+	if exp, got := -math.MaxFloat64, fields["value"].(float64); exp != got {
+		t.Fatalf("ParsePoints Value mismatch. \nexp: %v\ngot: %v", exp, got)
+	}
 
 	// leading zeros
 	_, err = models.ParsePointsString(fmt.Sprintf(`cpu,host=serverA,region=us-west value=%s`, "-0000000"+string(minFloat64)[1:]))
@@ -619,6 +682,61 @@ func TestParsePointMinFloat64(t *testing.T) {
 	}
 }
 
+func TestParsePointMaxUint64(t *testing.T) {
+	// out of range
+	_, err := models.ParsePointsString(`cpu,host=serverA,region=us-west value=18446744073709551616u`)
+	exp := `unable to parse 'cpu,host=serverA,region=us-west value=18446744073709551616u': unable to parse unsigned 18446744073709551616: strconv.ParseUint: parsing "18446744073709551616": value out of range`
+	if err == nil || (err != nil && err.Error() != exp) {
+		t.Fatalf("Error mismatch:\nexp: %s\ngot: %v", exp, err)
+	}
+
+	// max int
+	p, err := models.ParsePointsString(`cpu,host=serverA,region=us-west value=18446744073709551615u`)
+	if err != nil {
+		t.Fatalf(`ParsePoints("%s") mismatch. got %v, exp nil`, `cpu,host=serverA,region=us-west value=18446744073709551615u`, err)
+	}
+	fields, err := p[0].Fields()
+	if err != nil {
+		t.Fatal(err)
+	}
+	if exp, got := uint64(18446744073709551615), fields["value"].(uint64); exp != got {
+		t.Fatalf("ParsePoints Value mismatch. \nexp: %v\ngot: %v", exp, got)
+	}
+
+	// leading zeros
+	_, err = models.ParsePointsString(`cpu,host=serverA,region=us-west value=00018446744073709551615u`)
+	if err != nil {
+		t.Fatalf(`ParsePoints("%s") mismatch. got %v, exp nil`, `cpu,host=serverA,region=us-west value=00018446744073709551615u`, err)
+	}
+}
+
+func TestParsePointMinUint64(t *testing.T) {
+	// out of range
+	_, err := models.ParsePointsString(`cpu,host=serverA,region=us-west value=--1u`)
+	if err == nil {
+		t.Errorf(`ParsePoints("%s") mismatch. got nil, exp error`, `cpu,host=serverA,region=us-west value=-1u`)
+	}
+
+	// min int
+	p, err := models.ParsePointsString(`cpu,host=serverA,region=us-west value=0u`)
+	if err != nil {
+		t.Errorf(`ParsePoints("%s") mismatch. got %v, exp nil`, `cpu,host=serverA,region=us-west value=0u`, err)
+	}
+	fields, err := p[0].Fields()
+	if err != nil {
+		t.Fatal(err)
+	}
+	if exp, got := uint64(0), fields["value"].(uint64); exp != got {
+		t.Fatalf("ParsePoints Value mismatch. \nexp: %v\ngot: %v", exp, got)
+	}
+
+	// leading zeros
+	_, err = models.ParsePointsString(`cpu,host=serverA,region=us-west value=0000u`)
+	if err != nil {
+		t.Errorf(`ParsePoints("%s") mismatch. got %v, exp nil`, `cpu,host=serverA,region=us-west value=0000u`, err)
+	}
+}
+
 func TestParsePointNumberNonNumeric(t *testing.T) {
 	_, err := models.ParsePointsString(`cpu,host=serverA,region=us-west value=.1a`)
 	if err == nil {
@@ -1223,6 +1341,15 @@ func TestParsePointQuotedTags(t *testing.T) {
 	)
 }
 
+func TestParsePoint_TrailingSlash(t *testing.T) {
+	_, err := models.ParsePointsString(`a v=1 0\`)
+	if err == nil {
+		t.Fatalf("ParsePoints failed: %v", err)
+	} else if !strings.Contains(err.Error(), "bad timestamp") {
+		t.Fatalf("ParsePoints unexpected error: %v", err)
+	}
+}
+
 func TestParsePointsUnbalancedQuotedTags(t *testing.T) {
 	pts, err := models.ParsePointsString("baz,mytag=\"a x=1 1441103862125\nbaz,mytag=a z=1 1441103862126")
 	if err != nil {
@@ -2148,6 +2275,8 @@ func toFields(fi models.FieldIterator) models.Fields {
 			v, err = fi.FloatValue()
 		case models.Integer:
 			v, err = fi.IntegerValue()
+		case models.Unsigned:
+			v, err = fi.UnsignedValue()
 		case models.String:
 			v = fi.StringValue()
 		case models.Boolean:
@@ -2173,7 +2302,7 @@ m v=42i
 m v="string"
 m v=true
 m v="string\"with\"escapes"
-m v=42i,f=42,g=42.314
+m v=42i,f=42,g=42.314,u=123u
 m a=2i,b=3i,c=true,d="stuff",e=-0.23,f=123.456
 `)
 
@@ -2251,3 +2380,8 @@ func BenchmarkEscapeString_QuotesAndBackslashes(b *testing.B) {
 		sink = [...]string{models.EscapeStringField(s1), models.EscapeStringField(s2)}
 	}
 }
+
+func init() {
+	// Force uint support to be enabled for testing.
+	models.EnableUintSupport()
+}
