@@ -1,20 +1,19 @@
 package meta_test
 
 import (
-	"encoding/json"
 	"io/ioutil"
-	"net"
 	"os"
 	"path"
 	"reflect"
 	"runtime"
+	"strings"
 	"testing"
 	"time"
 
 	"github.com/influxdata/influxdb"
 
-	"github.com/influxdata/influxdb/influxql"
 	"github.com/influxdata/influxdb/services/meta"
+	"github.com/influxdata/influxql"
 )
 
 func TestMetaClient_CreateDatabaseOnly(t *testing.T) {
@@ -38,12 +37,12 @@ func TestMetaClient_CreateDatabaseOnly(t *testing.T) {
 	}
 
 	// Make sure a default retention policy was created.
-	rp, err := c.RetentionPolicy("db0", "default")
+	rp, err := c.RetentionPolicy("db0", "autogen")
 	if err != nil {
 		t.Fatal(err)
 	} else if rp == nil {
 		t.Fatal("failed to create rp")
-	} else if exp, got := "default", rp.Name; exp != got {
+	} else if exp, got := "autogen", rp.Name; exp != got {
 		t.Fatalf("rp name wrong:\n\texp: %s\n\tgot: %s", exp, got)
 	}
 }
@@ -78,13 +77,21 @@ func TestMetaClient_CreateDatabaseWithRetentionPolicy(t *testing.T) {
 	defer os.RemoveAll(d)
 	defer c.Close()
 
-	rpi := meta.RetentionPolicyInfo{
-		Name:               "rp0",
-		Duration:           1 * time.Hour,
-		ReplicaN:           1,
-		ShardGroupDuration: 2 * time.Hour,
+	// Calling CreateDatabaseWithRetentionPolicy with a nil spec should return
+	// an error
+	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", nil); err == nil {
+		t.Fatal("expected error")
 	}
-	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", &rpi); err != nil {
+
+	duration := 1 * time.Hour
+	replicaN := 1
+	spec := meta.RetentionPolicySpec{
+		Name:               "rp0",
+		Duration:           &duration,
+		ReplicaN:           &replicaN,
+		ShardGroupDuration: 60 * time.Minute,
+	}
+	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", &spec); err != nil {
 		t.Fatal(err)
 	}
 
@@ -102,34 +109,109 @@ func TestMetaClient_CreateDatabaseWithRetentionPolicy(t *testing.T) {
 		t.Fatalf("rp duration wrong: %v", rp.Duration)
 	} else if rp.ReplicaN != 1 {
 		t.Fatalf("rp replication wrong: %d", rp.ReplicaN)
-	} else if rp.ShardGroupDuration != 2*time.Hour {
+	} else if rp.ShardGroupDuration != 60*time.Minute {
 		t.Fatalf("rp shard duration wrong: %v", rp.ShardGroupDuration)
 	}
 
 	// Recreating the exact same database with retention policy is not
 	// an error.
-	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", &rpi); err != nil {
+	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", &spec); err != nil {
 		t.Fatal(err)
 	}
 
+	// If create database is used by itself, no error should be returned and
+	// the default retention policy should not be changed.
+	if dbi, err := c.CreateDatabase("db0"); err != nil {
+		t.Fatalf("got %v, but expected %v", err, nil)
+	} else if dbi.DefaultRetentionPolicy != "rp0" {
+		t.Fatalf("got %v, but expected %v", dbi.DefaultRetentionPolicy, "rp0")
+	} else if got, exp := len(dbi.RetentionPolicies), 1; got != exp {
+		// Ensure no additional retention policies were created.
+		t.Fatalf("got %v, but expected %v", got, exp)
+	}
+}
+
+func TestMetaClient_CreateDatabaseWithRetentionPolicy_Conflict_Fields(t *testing.T) {
+	t.Parallel()
+
+	d, c := newClient()
+	defer os.RemoveAll(d)
+	defer c.Close()
+
+	duration := 1 * time.Hour
+	replicaN := 1
+	spec := meta.RetentionPolicySpec{
+		Name:               "rp0",
+		Duration:           &duration,
+		ReplicaN:           &replicaN,
+		ShardGroupDuration: 60 * time.Minute,
+	}
+	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", &spec); err != nil {
+		t.Fatal(err)
+	}
+
+	// If the rp's name is different, and error should be returned.
+	spec2 := spec
+	spec2.Name = spec.Name + "1"
+	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", &spec2); err != meta.ErrRetentionPolicyConflict {
+		t.Fatalf("got %v, but expected %v", err, meta.ErrRetentionPolicyConflict)
+	}
+
 	// If the rp's duration is different, an error should be returned.
-	rpi2 := rpi
-	rpi2.Duration = rpi.Duration + time.Minute
-	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", &rpi2); err != meta.ErrRetentionPolicyConflict {
+	spec2 = spec
+	duration2 := *spec.Duration + time.Minute
+	spec2.Duration = &duration2
+	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", &spec2); err != meta.ErrRetentionPolicyConflict {
 		t.Fatalf("got %v, but expected %v", err, meta.ErrRetentionPolicyConflict)
 	}
 
 	// If the rp's replica is different, an error should be returned.
-	rpi2 = rpi
-	rpi2.ReplicaN = rpi.ReplicaN + 1
-	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", &rpi2); err != meta.ErrRetentionPolicyConflict {
+	spec2 = spec
+	replica2 := *spec.ReplicaN + 1
+	spec2.ReplicaN = &replica2
+	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", &spec2); err != meta.ErrRetentionPolicyConflict {
 		t.Fatalf("got %v, but expected %v", err, meta.ErrRetentionPolicyConflict)
 	}
 
 	// If the rp's shard group duration is different, an error should be returned.
-	rpi2 = rpi
-	rpi2.ShardGroupDuration = rpi.ShardGroupDuration + time.Minute
-	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", &rpi2); err != meta.ErrRetentionPolicyConflict {
+	spec2 = spec
+	spec2.ShardGroupDuration = spec.ShardGroupDuration + time.Minute
+	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", &spec2); err != meta.ErrRetentionPolicyConflict {
+		t.Fatalf("got %v, but expected %v", err, meta.ErrRetentionPolicyConflict)
+	}
+}
+
+func TestMetaClient_CreateDatabaseWithRetentionPolicy_Conflict_NonDefault(t *testing.T) {
+	t.Parallel()
+
+	d, c := newClient()
+	defer os.RemoveAll(d)
+	defer c.Close()
+
+	duration := 1 * time.Hour
+	replicaN := 1
+	spec := meta.RetentionPolicySpec{
+		Name:               "rp0",
+		Duration:           &duration,
+		ReplicaN:           &replicaN,
+		ShardGroupDuration: 60 * time.Minute,
+	}
+
+	// Create a default retention policy.
+	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", &spec); err != nil {
+		t.Fatal(err)
+	}
+
+	// Let's create a non-default retention policy.
+	spec2 := spec
+	spec2.Name = "rp1"
+	if _, err := c.CreateRetentionPolicy("db0", &spec2, false); err != nil {
+		t.Fatal(err)
+	}
+
+	// If we try to create a database with the non-default retention policy then
+	// it's an error.
+	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", &spec2); err != meta.ErrRetentionPolicyConflict {
 		t.Fatalf("got %v, but expected %v", err, meta.ErrRetentionPolicyConflict)
 	}
 }
@@ -224,11 +306,16 @@ func TestMetaClient_CreateRetentionPolicy(t *testing.T) {
 	rp0 := meta.RetentionPolicyInfo{
 		Name:               "rp0",
 		ReplicaN:           1,
-		Duration:           time.Hour,
-		ShardGroupDuration: time.Hour,
+		Duration:           2 * time.Hour,
+		ShardGroupDuration: 2 * time.Hour,
 	}
 
-	if _, err := c.CreateRetentionPolicy("db0", &rp0); err != nil {
+	if _, err := c.CreateRetentionPolicy("db0", &meta.RetentionPolicySpec{
+		Name:               rp0.Name,
+		ReplicaN:           &rp0.ReplicaN,
+		Duration:           &rp0.Duration,
+		ShardGroupDuration: rp0.ShardGroupDuration,
+	}, true); err != nil {
 		t.Fatal(err)
 	}
 
@@ -240,7 +327,12 @@ func TestMetaClient_CreateRetentionPolicy(t *testing.T) {
 	}
 
 	// Create the same policy.  Should not error.
-	if _, err := c.CreateRetentionPolicy("db0", &rp0); err != nil {
+	if _, err := c.CreateRetentionPolicy("db0", &meta.RetentionPolicySpec{
+		Name:               rp0.Name,
+		ReplicaN:           &rp0.ReplicaN,
+		Duration:           &rp0.Duration,
+		ShardGroupDuration: rp0.ShardGroupDuration,
+	}, true); err != nil {
 		t.Fatal(err)
 	} else if actual, err = c.RetentionPolicy("db0", "rp0"); err != nil {
 		t.Fatal(err)
@@ -253,7 +345,12 @@ func TestMetaClient_CreateRetentionPolicy(t *testing.T) {
 	rp1 := rp0
 	rp1.Duration = 2 * rp0.Duration
 
-	_, got := c.CreateRetentionPolicy("db0", &rp1)
+	_, got := c.CreateRetentionPolicy("db0", &meta.RetentionPolicySpec{
+		Name:               rp1.Name,
+		ReplicaN:           &rp1.ReplicaN,
+		Duration:           &rp1.Duration,
+		ShardGroupDuration: rp1.ShardGroupDuration,
+	}, true)
 	if exp := meta.ErrRetentionPolicyExists; got != exp {
 		t.Fatalf("got error %v, expected error %v", got, exp)
 	}
@@ -263,7 +360,12 @@ func TestMetaClient_CreateRetentionPolicy(t *testing.T) {
 	rp1 = rp0
 	rp1.ReplicaN = rp0.ReplicaN + 1
 
-	_, got = c.CreateRetentionPolicy("db0", &rp1)
+	_, got = c.CreateRetentionPolicy("db0", &meta.RetentionPolicySpec{
+		Name:               rp1.Name,
+		ReplicaN:           &rp1.ReplicaN,
+		Duration:           &rp1.Duration,
+		ShardGroupDuration: rp1.ShardGroupDuration,
+	}, true)
 	if exp := meta.ErrRetentionPolicyExists; got != exp {
 		t.Fatalf("got error %v, expected error %v", got, exp)
 	}
@@ -271,25 +373,48 @@ func TestMetaClient_CreateRetentionPolicy(t *testing.T) {
 	// Creating the same policy, but with a different shard group
 	// duration should also result in an error.
 	rp1 = rp0
-	rp1.ShardGroupDuration = 2 * rp0.ShardGroupDuration
+	rp1.ShardGroupDuration = rp0.ShardGroupDuration / 2
 
-	_, got = c.CreateRetentionPolicy("db0", &rp1)
+	_, got = c.CreateRetentionPolicy("db0", &meta.RetentionPolicySpec{
+		Name:               rp1.Name,
+		ReplicaN:           &rp1.ReplicaN,
+		Duration:           &rp1.Duration,
+		ShardGroupDuration: rp1.ShardGroupDuration,
+	}, true)
 	if exp := meta.ErrRetentionPolicyExists; got != exp {
 		t.Fatalf("got error %v, expected error %v", got, exp)
 	}
+
+	// Creating a policy with the shard duration being greater than the
+	// duration should also be an error.
+	rp1 = rp0
+	rp1.Duration = 1 * time.Hour
+	rp1.ShardGroupDuration = 2 * time.Hour
+
+	_, got = c.CreateRetentionPolicy("db0", &meta.RetentionPolicySpec{
+		Name:               rp1.Name,
+		ReplicaN:           &rp1.ReplicaN,
+		Duration:           &rp1.Duration,
+		ShardGroupDuration: rp1.ShardGroupDuration,
+	}, true)
+	if exp := meta.ErrIncompatibleDurations; got != exp {
+		t.Fatalf("got error %v, expected error %v", got, exp)
+	}
 }
 
-func TestMetaClient_SetDefaultRetentionPolicy(t *testing.T) {
+func TestMetaClient_DefaultRetentionPolicy(t *testing.T) {
 	t.Parallel()
 
 	d, c := newClient()
 	defer os.RemoveAll(d)
 	defer c.Close()
 
-	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", &meta.RetentionPolicyInfo{
+	duration := 1 * time.Hour
+	replicaN := 1
+	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", &meta.RetentionPolicySpec{
 		Name:     "rp0",
-		Duration: 1 * time.Hour,
-		ReplicaN: 1,
+		Duration: &duration,
+		ReplicaN: &replicaN,
 	}); err != nil {
 		t.Fatal(err)
 	}
@@ -318,6 +443,87 @@ func TestMetaClient_SetDefaultRetentionPolicy(t *testing.T) {
 	}
 }
 
+func TestMetaClient_UpdateRetentionPolicy(t *testing.T) {
+	t.Parallel()
+
+	d, c := newClient()
+	defer os.RemoveAll(d)
+	defer c.Close()
+
+	if _, err := c.CreateDatabaseWithRetentionPolicy("db0", &meta.RetentionPolicySpec{
+		Name:               "rp0",
+		ShardGroupDuration: 4 * time.Hour,
+	}); err != nil {
+		t.Fatal(err)
+	}
+
+	rpi, err := c.RetentionPolicy("db0", "rp0")
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	// Set the duration to another value and ensure that the shard group duration
+	// doesn't change.
+	duration := 2 * rpi.ShardGroupDuration
+	replicaN := 1
+	if err := c.UpdateRetentionPolicy("db0", "rp0", &meta.RetentionPolicyUpdate{
+		Duration: &duration,
+		ReplicaN: &replicaN,
+	}, true); err != nil {
+		t.Fatal(err)
+	}
+
+	rpi, err = c.RetentionPolicy("db0", "rp0")
+	if err != nil {
+		t.Fatal(err)
+	}
+	if exp, got := 4*time.Hour, rpi.ShardGroupDuration; exp != got {
+		t.Fatalf("shard group duration wrong: \n\texp: %s\n\tgot: %s", exp, got)
+	}
+
+	// Set the duration to below the shard group duration. This should return an error.
+	duration = rpi.ShardGroupDuration / 2
+	if err := c.UpdateRetentionPolicy("db0", "rp0", &meta.RetentionPolicyUpdate{
+		Duration: &duration,
+	}, true); err == nil {
+		t.Fatal("expected error")
+	} else if err != meta.ErrIncompatibleDurations {
+		t.Fatalf("expected error '%s', got '%s'", meta.ErrIncompatibleDurations, err)
+	}
+
+	// Set the shard duration longer than the overall duration. This should also return an error.
+	sgDuration := rpi.Duration * 2
+	if err := c.UpdateRetentionPolicy("db0", "rp0", &meta.RetentionPolicyUpdate{
+		ShardGroupDuration: &sgDuration,
+	}, true); err == nil {
+		t.Fatal("expected error")
+	} else if err != meta.ErrIncompatibleDurations {
+		t.Fatalf("expected error '%s', got '%s'", meta.ErrIncompatibleDurations, err)
+	}
+
+	// Set both values to incompatible values and ensure an error is returned.
+	duration = rpi.ShardGroupDuration
+	sgDuration = rpi.Duration
+	if err := c.UpdateRetentionPolicy("db0", "rp0", &meta.RetentionPolicyUpdate{
+		Duration:           &duration,
+		ShardGroupDuration: &sgDuration,
+	}, true); err == nil {
+		t.Fatal("expected error")
+	} else if err != meta.ErrIncompatibleDurations {
+		t.Fatalf("expected error '%s', got '%s'", meta.ErrIncompatibleDurations, err)
+	}
+
+	// Allow any shard duration if the duration is set to zero.
+	duration = time.Duration(0)
+	sgDuration = 168 * time.Hour
+	if err := c.UpdateRetentionPolicy("db0", "rp0", &meta.RetentionPolicyUpdate{
+		Duration:           &duration,
+		ShardGroupDuration: &sgDuration,
+	}, true); err != nil {
+		t.Fatalf("unexpected error: %v", err)
+	}
+}
+
 func TestMetaClient_DropRetentionPolicy(t *testing.T) {
 	t.Parallel()
 
@@ -336,11 +542,13 @@ func TestMetaClient_DropRetentionPolicy(t *testing.T) {
 		t.Fatalf("db name wrong: %s", db.Name)
 	}
 
-	if _, err := c.CreateRetentionPolicy("db0", &meta.RetentionPolicyInfo{
+	duration := 1 * time.Hour
+	replicaN := 1
+	if _, err := c.CreateRetentionPolicy("db0", &meta.RetentionPolicySpec{
 		Name:     "rp0",
-		Duration: 1 * time.Hour,
-		ReplicaN: 1,
-	}); err != nil {
+		Duration: &duration,
+		ReplicaN: &replicaN,
+	}, true); err != nil {
 		t.Fatal(err)
 	}
 
@@ -388,15 +596,15 @@ func TestMetaClient_CreateUser(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
-	if exp, got := "fred", u.Name; exp != got {
+	if exp, got := "fred", u.ID(); exp != got {
 		t.Fatalf("unexpected user name: exp: %s got: %s", exp, got)
 	}
-	if !u.Admin {
+	if !u.IsAdmin() {
 		t.Fatalf("expected user to be admin")
 	}
 
 	u, err = c.Authenticate("fred", "supersecure")
-	if u == nil || err != nil || u.Name != "fred" {
+	if u == nil || err != nil || u.ID() != "fred" {
 		t.Fatalf("failed to authenticate")
 	}
 
@@ -425,7 +633,7 @@ func TestMetaClient_CreateUser(t *testing.T) {
 
 	// Auth for new password should succeed.
 	u, err = c.Authenticate("fred", "moresupersecure")
-	if u == nil || err != nil || u.Name != "fred" {
+	if u == nil || err != nil || u.ID() != "fred" {
 		t.Fatalf("failed to authenticate")
 	}
 
@@ -439,10 +647,10 @@ func TestMetaClient_CreateUser(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
-	if exp, got := "wilma", u.Name; exp != got {
+	if exp, got := "wilma", u.ID(); exp != got {
 		t.Fatalf("unexpected user name: exp: %s got: %s", exp, got)
 	}
-	if u.Admin {
+	if u.IsAdmin() {
 		t.Fatalf("expected user not to be an admin")
 	}
 
@@ -459,10 +667,10 @@ func TestMetaClient_CreateUser(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
-	if exp, got := "wilma", u.Name; exp != got {
+	if exp, got := "wilma", u.ID(); exp != got {
 		t.Fatalf("unexpected user name: exp: %s got: %s", exp, got)
 	}
-	if !u.Admin {
+	if !u.IsAdmin() {
 		t.Fatalf("expected user to be an admin")
 	}
 
@@ -475,10 +683,10 @@ func TestMetaClient_CreateUser(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
-	if exp, got := "wilma", u.Name; exp != got {
+	if exp, got := "wilma", u.ID(); exp != got {
 		t.Fatalf("unexpected user name: exp: %s got: %s", exp, got)
 	}
-	if u.Admin {
+	if u.IsAdmin() {
 		t.Fatalf("expected user not to be an admin")
 	}
 
@@ -538,6 +746,19 @@ func TestMetaClient_CreateUser(t *testing.T) {
 	}
 }
 
+func TestMetaClient_UpdateUser(t *testing.T) {
+	t.Parallel()
+
+	d, c := newClient()
+	defer os.RemoveAll(d)
+	defer c.Close()
+
+	// UpdateUser that doesn't exist should return an error.
+	if err := c.UpdateUser("foo", "bar"); err == nil {
+		t.Fatalf("expected error, got nil")
+	}
+}
+
 func TestMetaClient_ContinuousQueries(t *testing.T) {
 	t.Parallel()
 
@@ -587,6 +808,11 @@ func TestMetaClient_ContinuousQueries(t *testing.T) {
 	if err := c.DropContinuousQuery("db0", "cq1"); err != nil {
 		t.Fatal(err)
 	}
+
+	// Dropping a nonexistent CQ should return an error.
+	if err := c.DropContinuousQuery("db0", "not-a-cq"); err == nil {
+		t.Fatal("expected an error, got nil")
+	}
 }
 
 func TestMetaClient_Subscriptions_Create(t *testing.T) {
@@ -608,17 +834,40 @@ func TestMetaClient_Subscriptions_Create(t *testing.T) {
 	}
 
 	// Create a subscription
-	if err := c.CreateSubscription("db0", "default", "sub0", "ALL", []string{"udp://example.com:9090"}); err != nil {
+	if err := c.CreateSubscription("db0", "autogen", "sub0", "ALL", []string{"udp://example.com:9090"}); err != nil {
 		t.Fatal(err)
 	}
 
 	// Re-create a subscription
-	if err := c.CreateSubscription("db0", "default", "sub0", "ALL", []string{"udp://example.com:9090"}); err == nil || err.Error() != `subscription already exists` {
+	err := c.CreateSubscription("db0", "autogen", "sub0", "ALL", []string{"udp://example.com:9090"})
+	if err == nil || err.Error() != `subscription already exists` {
 		t.Fatalf("unexpected error: %s", err)
 	}
 
 	// Create another subscription.
-	if err := c.CreateSubscription("db0", "default", "sub1", "ALL", []string{"udp://example.com:6060"}); err != nil {
+	if err := c.CreateSubscription("db0", "autogen", "sub1", "ALL", []string{"udp://example.com:6060"}); err != nil {
+		t.Fatal(err)
+	}
+
+	// Create a subscription with invalid scheme
+	err = c.CreateSubscription("db0", "autogen", "sub2", "ALL", []string{"bad://example.com:9191"})
+	if err == nil || !strings.HasPrefix(err.Error(), "invalid subscription URL") {
+		t.Fatalf("unexpected error: %s", err)
+	}
+
+	// Create a subscription without port number
+	err = c.CreateSubscription("db0", "autogen", "sub2", "ALL", []string{"udp://example.com"})
+	if err == nil || !strings.HasPrefix(err.Error(), "invalid subscription URL") {
+		t.Fatalf("unexpected error: %s", err)
+	}
+
+	// Create an HTTP subscription.
+	if err := c.CreateSubscription("db0", "autogen", "sub3", "ALL", []string{"http://example.com:9092"}); err != nil {
+		t.Fatal(err)
+	}
+
+	// Create an HTTPS subscription.
+	if err := c.CreateSubscription("db0", "autogen", "sub4", "ALL", []string{"https://example.com:9092"}); err != nil {
 		t.Fatal(err)
 	}
 }
@@ -637,19 +886,19 @@ func TestMetaClient_Subscriptions_Drop(t *testing.T) {
 
 	// DROP SUBSCRIPTION returns ErrSubscriptionNotFound when the
 	// subscription is unknown.
-	err := c.DropSubscription("db0", "default", "foo")
+	err := c.DropSubscription("db0", "autogen", "foo")
 	if got, exp := err, meta.ErrSubscriptionNotFound; got == nil || got.Error() != exp.Error() {
 		t.Fatalf("got: %s, exp: %s", got, exp)
 	}
 
 	// Create a subscription.
-	if err := c.CreateSubscription("db0", "default", "sub0", "ALL", []string{"udp://example.com:9090"}); err != nil {
+	if err := c.CreateSubscription("db0", "autogen", "sub0", "ALL", []string{"udp://example.com:9090"}); err != nil {
 		t.Fatal(err)
 	}
 
 	// DROP SUBSCRIPTION returns an influxdb.ErrDatabaseNotFound when
 	// the database is unknown.
-	err = c.DropSubscription("foo", "default", "sub0")
+	err = c.DropSubscription("foo", "autogen", "sub0")
 	if got, exp := err, influxdb.ErrDatabaseNotFound("foo"); got.Error() != exp.Error() {
 		t.Fatalf("got: %s, exp: %s", got, exp)
 	}
@@ -662,7 +911,7 @@ func TestMetaClient_Subscriptions_Drop(t *testing.T) {
 	}
 
 	// DROP SUBSCRIPTION drops the subsciption if it can find it.
-	err = c.DropSubscription("db0", "default", "sub0")
+	err = c.DropSubscription("db0", "autogen", "sub0")
 	if got := err; got != nil {
 		t.Fatalf("got: %s, exp: %v", got, nil)
 	}
@@ -681,7 +930,7 @@ func TestMetaClient_Shards(t *testing.T) {
 
 	// Test creating a shard group.
 	tmin := time.Now()
-	sg, err := c.CreateShardGroup("db0", "default", tmin)
+	sg, err := c.CreateShardGroup("db0", "autogen", tmin)
 	if err != nil {
 		t.Fatal(err)
 	} else if sg == nil {
@@ -696,7 +945,7 @@ func TestMetaClient_Shards(t *testing.T) {
 	}
 
 	// Test finding shard groups by time range.
-	groups, err := c.ShardGroupsByTimeRange("db0", "default", tmin, tmax)
+	groups, err := c.ShardGroupsByTimeRange("db0", "autogen", tmin, tmax)
 	if err != nil {
 		t.Fatal(err)
 	} else if len(groups) != 2 {
@@ -707,16 +956,16 @@ func TestMetaClient_Shards(t *testing.T) {
 	db, rp, owner := c.ShardOwner(groups[0].Shards[0].ID)
 	if db != "db0" {
 		t.Fatalf("wrong db name: %s", db)
-	} else if rp != "default" {
+	} else if rp != "autogen" {
 		t.Fatalf("wrong rp name: %s", rp)
 	} else if owner.ID != groups[0].ID {
 		t.Fatalf("wrong owner: exp %d got %d", groups[0].ID, owner.ID)
 	}
 
 	// Test deleting a shard group.
-	if err := c.DeleteShardGroup("db0", "default", groups[0].ID); err != nil {
+	if err := c.DeleteShardGroup("db0", "autogen", groups[0].ID); err != nil {
 		t.Fatal(err)
-	} else if groups, err = c.ShardGroupsByTimeRange("db0", "default", tmin, tmax); err != nil {
+	} else if groups, err = c.ShardGroupsByTimeRange("db0", "autogen", tmin, tmax); err != nil {
 		t.Fatal(err)
 	} else if len(groups) != 1 {
 		t.Fatalf("wrong number of shard groups after delete: %d", len(groups))
@@ -737,7 +986,7 @@ func TestMetaClient_CreateShardGroupIdempotent(t *testing.T) {
 
 	// create a shard group.
 	tmin := time.Now()
-	sg, err := c.CreateShardGroup("db0", "default", tmin)
+	sg, err := c.CreateShardGroup("db0", "autogen", tmin)
 	if err != nil {
 		t.Fatal(err)
 	} else if sg == nil {
@@ -748,7 +997,7 @@ func TestMetaClient_CreateShardGroupIdempotent(t *testing.T) {
 	t.Log("index: ", i)
 
 	// create the same shard group.
-	sg, err = c.CreateShardGroup("db0", "default", tmin)
+	sg, err = c.CreateShardGroup("db0", "autogen", tmin)
 	if err != nil {
 		t.Fatal(err)
 	} else if sg == nil {
@@ -778,6 +1027,85 @@ func TestMetaClient_CreateShardGroupIdempotent(t *testing.T) {
 	}
 }
 
+func TestMetaClient_PruneShardGroups(t *testing.T) {
+	t.Parallel()
+
+	d, c := newClient()
+	defer os.RemoveAll(d)
+	defer c.Close()
+
+	if _, err := c.CreateDatabase("db0"); err != nil {
+		t.Fatal(err)
+	}
+
+	if _, err := c.CreateDatabase("db1"); err != nil {
+		t.Fatal(err)
+	}
+
+	duration := 1 * time.Hour
+	replicaN := 1
+
+	if _, err := c.CreateRetentionPolicy("db1", &meta.RetentionPolicySpec{
+		Name:     "rp0",
+		Duration: &duration,
+		ReplicaN: &replicaN,
+	}, true); err != nil {
+		t.Fatal(err)
+	}
+
+	sg, err := c.CreateShardGroup("db1", "autogen", time.Now())
+	if err != nil {
+		t.Fatal(err)
+	} else if sg == nil {
+		t.Fatalf("expected ShardGroup")
+	}
+
+	sg, err = c.CreateShardGroup("db1", "autogen", time.Now().Add(15*24*time.Hour))
+	if err != nil {
+		t.Fatal(err)
+	} else if sg == nil {
+		t.Fatalf("expected ShardGroup")
+	}
+
+	sg, err = c.CreateShardGroup("db1", "rp0", time.Now())
+	if err != nil {
+		t.Fatal(err)
+	} else if sg == nil {
+		t.Fatalf("expected ShardGroup")
+	}
+
+	expiration := time.Now().Add(-2 * 7 * 24 * time.Hour).Add(-1 * time.Hour)
+
+	data := c.Data()
+	data.Databases[1].RetentionPolicies[0].ShardGroups[0].DeletedAt = expiration
+	data.Databases[1].RetentionPolicies[0].ShardGroups[1].DeletedAt = expiration
+
+	if err := c.SetData(&data); err != nil {
+		t.Fatal(err)
+	}
+
+	if err := c.PruneShardGroups(); err != nil {
+		t.Fatal(err)
+	}
+
+	data = c.Data()
+	rp, err := data.RetentionPolicy("db1", "autogen")
+	if err != nil {
+		t.Fatal(err)
+	}
+	if got, exp := len(rp.ShardGroups), 0; got != exp {
+		t.Fatalf("failed to prune shard group. got: %d, exp: %d", got, exp)
+	}
+
+	rp, err = data.RetentionPolicy("db1", "rp0")
+	if err != nil {
+		t.Fatal(err)
+	}
+	if got, exp := len(rp.ShardGroups), 1; got != exp {
+		t.Fatalf("failed to prune shard group. got: %d, exp: %d", got, exp)
+	}
+}
+
 func TestMetaClient_PersistClusterIDAfterRestart(t *testing.T) {
 	t.Parallel()
 
@@ -836,33 +1164,3 @@ func testTempDir(skip int) string {
 	}
 	return dir
 }
-
-func mustParseStatement(s string) influxql.Statement {
-	stmt, err := influxql.ParseStatement(s)
-	if err != nil {
-		panic(err)
-	}
-	return stmt
-}
-
-func mustMarshalJSON(v interface{}) string {
-	b, e := json.Marshal(v)
-	if e != nil {
-		panic(e)
-	}
-	return string(b)
-}
-
-func freePort() string {
-	l, _ := net.Listen("tcp", "127.0.0.1:0")
-	defer l.Close()
-	return l.Addr().String()
-}
-
-func freePorts(i int) []string {
-	var ports []string
-	for j := 0; j < i; j++ {
-		ports = append(ports, freePort())
-	}
-	return ports
-}
