@@ -2,16 +2,20 @@ package meta
 
 import (
 	"errors"
-	"fmt"
+	"net"
+	"net/url"
 	"sort"
 	"strings"
 	"sync"
 	"time"
+	"unicode"
 
 	"github.com/gogo/protobuf/proto"
 	"github.com/influxdata/influxdb"
-	"github.com/influxdata/influxdb/influxql"
+	"github.com/influxdata/influxdb/models"
+	"github.com/influxdata/influxdb/query"
 	internal "github.com/influxdata/influxdb/services/meta/internal"
+	"github.com/influxdata/influxql"
 )
 
 //go:generate protoc --gogo_out=. internal/meta.proto
@@ -21,7 +25,10 @@ const (
 	DefaultRetentionPolicyReplicaN = 1
 
 	// DefaultRetentionPolicyDuration is the default value of RetentionPolicyInfo.Duration.
-	DefaultRetentionPolicyDuration = 7 * (24 * time.Hour)
+	DefaultRetentionPolicyDuration = time.Duration(0)
+
+	// DefaultRetentionPolicyName is the default name for auto generated retention policies.
+	DefaultRetentionPolicyName = "autogen"
 
 	// MinRetentionPolicyDuration represents the minimum duration for a policy.
 	MinRetentionPolicyDuration = time.Hour
@@ -35,37 +42,15 @@ type Data struct {
 	Databases []DatabaseInfo
 	Users     []UserInfo
 
+	// adminUserExists provides a constant time mechanism for determining
+	// if there is at least one admin user.
+	adminUserExists bool
+
 	MaxShardGroupID uint64
 	MaxShardID      uint64
 }
 
-// NewShardOwner sets the owner of the provided shard to the data node
-// that currently owns the fewest number of shards. If multiple nodes
-// own the same (fewest) number of shards, then one of those nodes
-// becomes the new shard owner.
-func NewShardOwner(s ShardInfo, ownerFreqs map[int]int) (uint64, error) {
-	var (
-		minId   = -1
-		minFreq int
-	)
-
-	for id, freq := range ownerFreqs {
-		if minId == -1 || freq < minFreq {
-			minId, minFreq = int(id), freq
-		}
-	}
-
-	if minId < 0 {
-		return 0, fmt.Errorf("cannot reassign shard %d due to lack of data nodes", s.ID)
-	}
-
-	// Update the shard owner frequencies and set the new owner on the
-	// shard.
-	ownerFreqs[minId]++
-	return uint64(minId), nil
-}
-
-// Database returns a database by name.
+// Database returns a DatabaseInfo by the database name.
 func (data *Data) Database(name string) *DatabaseInfo {
 	for i := range data.Databases {
 		if data.Databases[i].Name == name {
@@ -75,7 +60,7 @@ func (data *Data) Database(name string) *DatabaseInfo {
 	return nil
 }
 
-// CloneDatabases returns a copy of the databases.
+// CloneDatabases returns a copy of the DatabaseInfo.
 func (data *Data) CloneDatabases() []DatabaseInfo {
 	if data.Databases == nil {
 		return nil
@@ -88,7 +73,7 @@ func (data *Data) CloneDatabases() []DatabaseInfo {
 }
 
 // CreateDatabase creates a new database.
-// Returns an error if name is blank or if a database with the same name already exists.
+// It returns an error if name is blank or if a database with the same name already exists.
 func (data *Data) CreateDatabase(name string) error {
 	if name == "" {
 		return ErrDatabaseNameRequired
@@ -108,6 +93,11 @@ func (data *Data) DropDatabase(name string) error {
 	for i := range data.Databases {
 		if data.Databases[i].Name == name {
 			data.Databases = append(data.Databases[:i], data.Databases[i+1:]...)
+
+			// Remove all user privileges associated with this database.
+			for i := range data.Users {
+				delete(data.Users[i].Privileges, name)
+			}
 			break
 		}
 	}
@@ -130,39 +120,50 @@ func (data *Data) RetentionPolicy(database, name string) (*RetentionPolicyInfo,
 }
 
 // CreateRetentionPolicy creates a new retention policy on a database.
-// Returns an error if name is blank or if a database does not exist.
-func (data *Data) CreateRetentionPolicy(database string, rpi *RetentionPolicyInfo) error {
+// It returns an error if name is blank or if the database does not exist.
+func (data *Data) CreateRetentionPolicy(database string, rpi *RetentionPolicyInfo, makeDefault bool) error {
 	// Validate retention policy.
-	if rpi.Name == "" {
+	if rpi == nil {
+		return ErrRetentionPolicyRequired
+	} else if rpi.Name == "" {
 		return ErrRetentionPolicyNameRequired
 	} else if rpi.ReplicaN < 1 {
 		return ErrReplicationFactorTooLow
 	}
 
 	// Normalise ShardDuration before comparing to any existing
-	// retention policies
+	// retention policies. The client is supposed to do this, but
+	// do it again to verify input.
 	rpi.ShardGroupDuration = normalisedShardDuration(rpi.ShardGroupDuration, rpi.Duration)
 
+	if rpi.Duration > 0 && rpi.Duration < rpi.ShardGroupDuration {
+		return ErrIncompatibleDurations
+	}
+
 	// Find database.
 	di := data.Database(database)
 	if di == nil {
 		return influxdb.ErrDatabaseNotFound(database)
 	} else if rp := di.RetentionPolicy(rpi.Name); rp != nil {
-		// RP with that name already exists.  Make sure they're the same.
+		// RP with that name already exists. Make sure they're the same.
 		if rp.ReplicaN != rpi.ReplicaN || rp.Duration != rpi.Duration || rp.ShardGroupDuration != rpi.ShardGroupDuration {
 			return ErrRetentionPolicyExists
 		}
+		// if they want to make it default, and it's not the default, it's not an identical command so it's an error
+		if makeDefault && di.DefaultRetentionPolicy != rpi.Name {
+			return ErrRetentionPolicyConflict
+		}
 		return nil
 	}
 
 	// Append copy of new policy.
-	rp := RetentionPolicyInfo{
-		Name:               rpi.Name,
-		Duration:           rpi.Duration,
-		ReplicaN:           rpi.ReplicaN,
-		ShardGroupDuration: rpi.ShardGroupDuration,
+	di.RetentionPolicies = append(di.RetentionPolicies, *rpi)
+
+	// Set the default if needed
+	if makeDefault {
+		di.DefaultRetentionPolicy = rpi.Name
 	}
-	di.RetentionPolicies = append(di.RetentionPolicies, rp)
+
 	return nil
 }
 
@@ -194,20 +195,20 @@ type RetentionPolicyUpdate struct {
 	ShardGroupDuration *time.Duration
 }
 
-// SetName sets the RetentionPolicyUpdate.Name
+// SetName sets the RetentionPolicyUpdate.Name.
 func (rpu *RetentionPolicyUpdate) SetName(v string) { rpu.Name = &v }
 
-// SetDuration sets the RetentionPolicyUpdate.Duration
+// SetDuration sets the RetentionPolicyUpdate.Duration.
 func (rpu *RetentionPolicyUpdate) SetDuration(v time.Duration) { rpu.Duration = &v }
 
-// SetReplicaN sets the RetentionPolicyUpdate.ReplicaN
+// SetReplicaN sets the RetentionPolicyUpdate.ReplicaN.
 func (rpu *RetentionPolicyUpdate) SetReplicaN(v int) { rpu.ReplicaN = &v }
 
-// SetShardGroupDuration sets the RetentionPolicyUpdate.ShardGroupDuration
+// SetShardGroupDuration sets the RetentionPolicyUpdate.ShardGroupDuration.
 func (rpu *RetentionPolicyUpdate) SetShardGroupDuration(v time.Duration) { rpu.ShardGroupDuration = &v }
 
 // UpdateRetentionPolicy updates an existing retention policy.
-func (data *Data) UpdateRetentionPolicy(database, name string, rpu *RetentionPolicyUpdate) error {
+func (data *Data) UpdateRetentionPolicy(database, name string, rpu *RetentionPolicyUpdate, makeDefault bool) error {
 	// Find database.
 	di := data.Database(database)
 	if di == nil {
@@ -230,6 +231,15 @@ func (data *Data) UpdateRetentionPolicy(database, name string, rpu *RetentionPol
 		return ErrRetentionPolicyDurationTooLow
 	}
 
+	// Enforce duration is at least the shard duration
+	if (rpu.Duration != nil && *rpu.Duration > 0 &&
+		((rpu.ShardGroupDuration != nil && *rpu.Duration < *rpu.ShardGroupDuration) ||
+			(rpu.ShardGroupDuration == nil && *rpu.Duration < rpi.ShardGroupDuration))) ||
+		(rpu.Duration == nil && rpi.Duration > 0 &&
+			rpu.ShardGroupDuration != nil && rpi.Duration < *rpu.ShardGroupDuration) {
+		return ErrIncompatibleDurations
+	}
+
 	// Update fields.
 	if rpu.Name != nil {
 		rpi.Name = *rpu.Name
@@ -240,29 +250,14 @@ func (data *Data) UpdateRetentionPolicy(database, name string, rpu *RetentionPol
 	if rpu.ReplicaN != nil {
 		rpi.ReplicaN = *rpu.ReplicaN
 	}
-
 	if rpu.ShardGroupDuration != nil {
-		rpi.ShardGroupDuration = *rpu.ShardGroupDuration
-	} else {
-		rpi.ShardGroupDuration = shardGroupDuration(rpi.Duration)
+		rpi.ShardGroupDuration = normalisedShardDuration(*rpu.ShardGroupDuration, rpi.Duration)
 	}
 
-	return nil
-}
-
-// SetDefaultRetentionPolicy sets the default retention policy for a database.
-func (data *Data) SetDefaultRetentionPolicy(database, name string) error {
-	// Find database and verify policy exists.
-	di := data.Database(database)
-	if di == nil {
-		return influxdb.ErrDatabaseNotFound(database)
-	} else if di.RetentionPolicy(name) == nil {
-		return influxdb.ErrRetentionPolicyNotFound(name)
+	if di.DefaultRetentionPolicy != rpi.Name && makeDefault {
+		di.DefaultRetentionPolicy = rpi.Name
 	}
 
-	// Set default policy.
-	di.DefaultRetentionPolicy = name
-
 	return nil
 }
 
@@ -298,7 +293,7 @@ func (data *Data) DropShard(id uint64) {
 	}
 }
 
-// ShardGroups returns a list of all shard groups on a database and policy.
+// ShardGroups returns a list of all shard groups on a database and retention policy.
 func (data *Data) ShardGroups(database, policy string) ([]ShardGroupInfo, error) {
 	// Find retention policy.
 	rpi, err := data.RetentionPolicy(database, policy)
@@ -371,6 +366,10 @@ func (data *Data) CreateShardGroup(database, policy string, timestamp time.Time)
 	sgi.ID = data.MaxShardGroupID
 	sgi.StartTime = timestamp.Truncate(rpi.ShardGroupDuration).UTC()
 	sgi.EndTime = sgi.StartTime.Add(rpi.ShardGroupDuration).UTC()
+	if sgi.EndTime.After(time.Unix(0, models.MaxNanoTime)) {
+		// Shard group range is [start, end) so add one to the max time.
+		sgi.EndTime = time.Unix(0, models.MaxNanoTime+1)
+	}
 
 	data.MaxShardID++
 	sgi.Shards = []ShardInfo{
@@ -452,8 +451,33 @@ func (data *Data) DropContinuousQuery(database, name string) error {
 	return ErrContinuousQueryNotFound
 }
 
+// validateURL returns an error if the URL does not have a port or uses a scheme other than UDP or HTTP.
+func validateURL(input string) error {
+	u, err := url.Parse(input)
+	if err != nil {
+		return ErrInvalidSubscriptionURL(input)
+	}
+
+	if u.Scheme != "udp" && u.Scheme != "http" && u.Scheme != "https" {
+		return ErrInvalidSubscriptionURL(input)
+	}
+
+	_, port, err := net.SplitHostPort(u.Host)
+	if err != nil || port == "" {
+		return ErrInvalidSubscriptionURL(input)
+	}
+
+	return nil
+}
+
 // CreateSubscription adds a named subscription to a database and retention policy.
 func (data *Data) CreateSubscription(database, rp, name, mode string, destinations []string) error {
+	for _, d := range destinations {
+		if err := validateURL(d); err != nil {
+			return err
+		}
+	}
+
 	rpi, err := data.RetentionPolicy(database, rp)
 	if err != nil {
 		return err
@@ -496,8 +520,7 @@ func (data *Data) DropSubscription(database, rp, name string) error {
 	return ErrSubscriptionNotFound
 }
 
-// User returns a user by username.
-func (data *Data) User(username string) *UserInfo {
+func (data *Data) user(username string) *UserInfo {
 	for i := range data.Users {
 		if data.Users[i].Name == username {
 			return &data.Users[i]
@@ -506,6 +529,16 @@ func (data *Data) User(username string) *UserInfo {
 	return nil
 }
 
+// User returns a user by username.
+func (data *Data) User(username string) User {
+	u := data.user(username)
+	if u == nil {
+		// prevent non-nil interface with nil pointer
+		return nil
+	}
+	return u
+}
+
 // CreateUser creates a new user.
 func (data *Data) CreateUser(name, hash string, admin bool) error {
 	// Ensure the user doesn't already exist.
@@ -522,6 +555,11 @@ func (data *Data) CreateUser(name, hash string, admin bool) error {
 		Admin: admin,
 	})
 
+	// We know there is now at least one admin user.
+	if admin {
+		data.adminUserExists = true
+	}
+
 	return nil
 }
 
@@ -529,10 +567,17 @@ func (data *Data) CreateUser(name, hash string, admin bool) error {
 func (data *Data) DropUser(name string) error {
 	for i := range data.Users {
 		if data.Users[i].Name == name {
+			wasAdmin := data.Users[i].Admin
 			data.Users = append(data.Users[:i], data.Users[i+1:]...)
+
+			// Maybe we dropped the only admin user?
+			if wasAdmin {
+				data.adminUserExists = data.hasAdminUser()
+			}
 			return nil
 		}
 	}
+
 	return ErrUserNotFound
 }
 
@@ -547,7 +592,7 @@ func (data *Data) UpdateUser(name, hash string) error {
 	return ErrUserNotFound
 }
 
-// CloneUsers returns a copy of the user infos
+// CloneUsers returns a copy of the user infos.
 func (data *Data) CloneUsers() []UserInfo {
 	if len(data.Users) == 0 {
 		return []UserInfo{}
@@ -562,11 +607,15 @@ func (data *Data) CloneUsers() []UserInfo {
 
 // SetPrivilege sets a privilege for a user on a database.
 func (data *Data) SetPrivilege(name, database string, p influxql.Privilege) error {
-	ui := data.User(name)
+	ui := data.user(name)
 	if ui == nil {
 		return ErrUserNotFound
 	}
 
+	if data.Database(database) == nil {
+		return influxdb.ErrDatabaseNotFound(database)
+	}
+
 	if ui.Privileges == nil {
 		ui.Privileges = make(map[string]influxql.Privilege)
 	}
@@ -577,19 +626,27 @@ func (data *Data) SetPrivilege(name, database string, p influxql.Privilege) erro
 
 // SetAdminPrivilege sets the admin privilege for a user.
 func (data *Data) SetAdminPrivilege(name string, admin bool) error {
-	ui := data.User(name)
+	ui := data.user(name)
 	if ui == nil {
 		return ErrUserNotFound
 	}
 
 	ui.Admin = admin
 
+	// We could have promoted or revoked the only admin. Check if an admin
+	// user exists.
+	data.adminUserExists = data.hasAdminUser()
 	return nil
 }
 
+// AdminUserExists returns true if an admin user exists.
+func (data Data) AdminUserExists() bool {
+	return data.adminUserExists
+}
+
 // UserPrivileges gets the privileges for a user.
 func (data *Data) UserPrivileges(name string) (map[string]influxql.Privilege, error) {
-	ui := data.User(name)
+	ui := data.user(name)
 	if ui == nil {
 		return nil, ErrUserNotFound
 	}
@@ -599,7 +656,7 @@ func (data *Data) UserPrivileges(name string) (map[string]influxql.Privilege, er
 
 // UserPrivilege gets the privilege for a user on a database.
 func (data *Data) UserPrivilege(name, database string) (*influxql.Privilege, error) {
-	ui := data.User(name)
+	ui := data.user(name)
 	if ui == nil {
 		return nil, ErrUserNotFound
 	}
@@ -623,7 +680,7 @@ func (data *Data) Clone() *Data {
 	return &other
 }
 
-// marshal serializes to a protobuf representation.
+// marshal serializes data to a protobuf representation.
 func (data *Data) marshal() *internal.Data {
 	pb := &internal.Data{
 		Term:      proto.Uint64(data.Term),
@@ -668,6 +725,10 @@ func (data *Data) unmarshal(pb *internal.Data) {
 	for i, x := range pb.GetUsers() {
 		data.Users[i].unmarshal(x)
 	}
+
+	// Exhaustively determine if there is an admin user. The marshalled cache
+	// value may not be correct.
+	data.adminUserExists = data.hasAdminUser()
 }
 
 // MarshalBinary encodes the metadata to a binary format.
@@ -685,6 +746,17 @@ func (data *Data) UnmarshalBinary(buf []byte) error {
 	return nil
 }
 
+// hasAdminUser exhaustively checks for the presence of at least one admin
+// user.
+func (data *Data) hasAdminUser() bool {
+	for _, u := range data.Users {
+		if u.Admin {
+			return true
+		}
+	}
+	return false
+}
+
 // NodeInfo represents information about a single node in the cluster.
 type NodeInfo struct {
 	ID      uint64
@@ -692,30 +764,16 @@ type NodeInfo struct {
 	TCPHost string
 }
 
-// clone returns a deep copy of ni.
-func (ni NodeInfo) clone() NodeInfo { return ni }
-
-// marshal serializes to a protobuf representation.
-func (ni NodeInfo) marshal() *internal.NodeInfo {
-	pb := &internal.NodeInfo{}
-	pb.ID = proto.Uint64(ni.ID)
-	pb.Host = proto.String(ni.Host)
-	pb.TCPHost = proto.String(ni.TCPHost)
-	return pb
-}
-
-// unmarshal deserializes from a protobuf representation.
-func (ni *NodeInfo) unmarshal(pb *internal.NodeInfo) {
-	ni.ID = pb.GetID()
-	ni.Host = pb.GetHost()
-	ni.TCPHost = pb.GetTCPHost()
-}
-
 // NodeInfos is a slice of NodeInfo used for sorting
 type NodeInfos []NodeInfo
 
-func (n NodeInfos) Len() int           { return len(n) }
-func (n NodeInfos) Swap(i, j int)      { n[i], n[j] = n[j], n[i] }
+// Len implements sort.Interface.
+func (n NodeInfos) Len() int { return len(n) }
+
+// Swap implements sort.Interface.
+func (n NodeInfos) Swap(i, j int) { n[i], n[j] = n[j], n[i] }
+
+// Less implements sort.Interface.
 func (n NodeInfos) Less(i, j int) bool { return n[i].ID < n[j].ID }
 
 // DatabaseInfo represents information about a database in the system.
@@ -728,6 +786,13 @@ type DatabaseInfo struct {
 
 // RetentionPolicy returns a retention policy by name.
 func (di DatabaseInfo) RetentionPolicy(name string) *RetentionPolicyInfo {
+	if name == "" {
+		if di.DefaultRetentionPolicy == "" {
+			return nil
+		}
+		name = di.DefaultRetentionPolicy
+	}
+
 	for i := range di.RetentionPolicies {
 		if di.RetentionPolicies[i].Name == name {
 			return &di.RetentionPolicies[i]
@@ -821,6 +886,90 @@ func (di *DatabaseInfo) unmarshal(pb *internal.DatabaseInfo) {
 	}
 }
 
+// RetentionPolicySpec represents the specification for a new retention policy.
+type RetentionPolicySpec struct {
+	Name               string
+	ReplicaN           *int
+	Duration           *time.Duration
+	ShardGroupDuration time.Duration
+}
+
+// NewRetentionPolicyInfo creates a new retention policy info from the specification.
+func (s *RetentionPolicySpec) NewRetentionPolicyInfo() *RetentionPolicyInfo {
+	return DefaultRetentionPolicyInfo().Apply(s)
+}
+
+// Matches checks if this retention policy specification matches
+// an existing retention policy.
+func (s *RetentionPolicySpec) Matches(rpi *RetentionPolicyInfo) bool {
+	if rpi == nil {
+		return false
+	} else if s.Name != "" && s.Name != rpi.Name {
+		return false
+	} else if s.Duration != nil && *s.Duration != rpi.Duration {
+		return false
+	} else if s.ReplicaN != nil && *s.ReplicaN != rpi.ReplicaN {
+		return false
+	}
+
+	// Normalise ShardDuration before comparing to any existing retention policies.
+	// Normalize with the retention policy info's duration instead of the spec
+	// since they should be the same and we're performing a comparison.
+	sgDuration := normalisedShardDuration(s.ShardGroupDuration, rpi.Duration)
+	return sgDuration == rpi.ShardGroupDuration
+}
+
+// marshal serializes to a protobuf representation.
+func (s *RetentionPolicySpec) marshal() *internal.RetentionPolicySpec {
+	pb := &internal.RetentionPolicySpec{}
+	if s.Name != "" {
+		pb.Name = proto.String(s.Name)
+	}
+	if s.Duration != nil {
+		pb.Duration = proto.Int64(int64(*s.Duration))
+	}
+	if s.ShardGroupDuration > 0 {
+		pb.ShardGroupDuration = proto.Int64(int64(s.ShardGroupDuration))
+	}
+	if s.ReplicaN != nil {
+		pb.ReplicaN = proto.Uint32(uint32(*s.ReplicaN))
+	}
+	return pb
+}
+
+// unmarshal deserializes from a protobuf representation.
+func (s *RetentionPolicySpec) unmarshal(pb *internal.RetentionPolicySpec) {
+	if pb.Name != nil {
+		s.Name = pb.GetName()
+	}
+	if pb.Duration != nil {
+		duration := time.Duration(pb.GetDuration())
+		s.Duration = &duration
+	}
+	if pb.ShardGroupDuration != nil {
+		s.ShardGroupDuration = time.Duration(pb.GetShardGroupDuration())
+	}
+	if pb.ReplicaN != nil {
+		replicaN := int(pb.GetReplicaN())
+		s.ReplicaN = &replicaN
+	}
+}
+
+// MarshalBinary encodes RetentionPolicySpec to a binary format.
+func (s *RetentionPolicySpec) MarshalBinary() ([]byte, error) {
+	return proto.Marshal(s.marshal())
+}
+
+// UnmarshalBinary decodes RetentionPolicySpec from a binary format.
+func (s *RetentionPolicySpec) UnmarshalBinary(data []byte) error {
+	var pb internal.RetentionPolicySpec
+	if err := proto.Unmarshal(data, &pb); err != nil {
+		return err
+	}
+	s.unmarshal(&pb)
+	return nil
+}
+
 // RetentionPolicyInfo represents metadata about a retention policy.
 type RetentionPolicyInfo struct {
 	Name               string
@@ -831,7 +980,8 @@ type RetentionPolicyInfo struct {
 	Subscriptions      []SubscriptionInfo
 }
 
-// NewRetentionPolicyInfo returns a new instance of RetentionPolicyInfo with defaults set.
+// NewRetentionPolicyInfo returns a new instance of RetentionPolicyInfo
+// with default replication and duration.
 func NewRetentionPolicyInfo(name string) *RetentionPolicyInfo {
 	return &RetentionPolicyInfo{
 		Name:     name,
@@ -840,13 +990,43 @@ func NewRetentionPolicyInfo(name string) *RetentionPolicyInfo {
 	}
 }
 
-// ShardGroupByTimestamp returns the shard group in the policy that contains the timestamp.
+// DefaultRetentionPolicyInfo returns a new instance of RetentionPolicyInfo
+// with default name, replication, and duration.
+func DefaultRetentionPolicyInfo() *RetentionPolicyInfo {
+	return NewRetentionPolicyInfo(DefaultRetentionPolicyName)
+}
+
+// Apply applies a specification to the retention policy info.
+func (rpi *RetentionPolicyInfo) Apply(spec *RetentionPolicySpec) *RetentionPolicyInfo {
+	rp := &RetentionPolicyInfo{
+		Name:               rpi.Name,
+		ReplicaN:           rpi.ReplicaN,
+		Duration:           rpi.Duration,
+		ShardGroupDuration: rpi.ShardGroupDuration,
+	}
+	if spec.Name != "" {
+		rp.Name = spec.Name
+	}
+	if spec.ReplicaN != nil {
+		rp.ReplicaN = *spec.ReplicaN
+	}
+	if spec.Duration != nil {
+		rp.Duration = *spec.Duration
+	}
+	rp.ShardGroupDuration = normalisedShardDuration(spec.ShardGroupDuration, rp.Duration)
+	return rp
+}
+
+// ShardGroupByTimestamp returns the shard group in the policy that contains the timestamp,
+// or nil if no shard group matches.
 func (rpi *RetentionPolicyInfo) ShardGroupByTimestamp(timestamp time.Time) *ShardGroupInfo {
 	for i := range rpi.ShardGroups {
-		if rpi.ShardGroups[i].Contains(timestamp) && !rpi.ShardGroups[i].Deleted() {
+		sgi := &rpi.ShardGroups[i]
+		if sgi.Contains(timestamp) && !sgi.Deleted() && (!sgi.Truncated() || timestamp.Before(sgi.TruncatedAt)) {
 			return &rpi.ShardGroups[i]
 		}
 	}
+
 	return nil
 }
 
@@ -947,7 +1127,7 @@ func (rpi *RetentionPolicyInfo) UnmarshalBinary(data []byte) error {
 	return nil
 }
 
-// shardGroupDuration returns the duration for a shard group based on a policy duration.
+// shardGroupDuration returns the default duration for a shard group based on a policy duration.
 func shardGroupDuration(d time.Duration) time.Duration {
 	if d >= 180*24*time.Hour || d == 0 { // 6 months or 0
 		return 7 * 24 * time.Hour
@@ -959,9 +1139,15 @@ func shardGroupDuration(d time.Duration) time.Duration {
 
 // normalisedShardDuration returns normalised shard duration based on a policy duration.
 func normalisedShardDuration(sgd, d time.Duration) time.Duration {
+	// If it is zero, it likely wasn't specified, so we default to the shard group duration
 	if sgd == 0 {
 		return shardGroupDuration(d)
 	}
+	// If it was specified, but it's less than the MinRetentionPolicyDuration, then normalize
+	// to the MinRetentionPolicyDuration
+	if sgd < MinRetentionPolicyDuration {
+		return shardGroupDuration(MinRetentionPolicyDuration)
+	}
 	return sgd
 }
 
@@ -970,27 +1156,49 @@ func normalisedShardDuration(sgd, d time.Duration) time.Duration {
 // to be sure that a ShardGroup is not simply missing. If the DeletedAt is set, the system can
 // safely delete any associated shards.
 type ShardGroupInfo struct {
-	ID        uint64
-	StartTime time.Time
-	EndTime   time.Time
-	DeletedAt time.Time
-	Shards    []ShardInfo
+	ID          uint64
+	StartTime   time.Time
+	EndTime     time.Time
+	DeletedAt   time.Time
+	Shards      []ShardInfo
+	TruncatedAt time.Time
 }
 
 // ShardGroupInfos implements sort.Interface on []ShardGroupInfo, based
 // on the StartTime field.
 type ShardGroupInfos []ShardGroupInfo
 
-func (a ShardGroupInfos) Len() int           { return len(a) }
-func (a ShardGroupInfos) Less(i, j int) bool { return a[i].StartTime.Before(a[j].StartTime) }
-func (a ShardGroupInfos) Swap(i, j int)      { a[i], a[j] = a[j], a[i] }
+// Len implements sort.Interface.
+func (a ShardGroupInfos) Len() int { return len(a) }
 
-// Contains return true if the shard group contains data for the timestamp.
+// Swap implements sort.Interface.
+func (a ShardGroupInfos) Swap(i, j int) { a[i], a[j] = a[j], a[i] }
+
+// Less implements sort.Interface.
+func (a ShardGroupInfos) Less(i, j int) bool {
+	iEnd := a[i].EndTime
+	if a[i].Truncated() {
+		iEnd = a[i].TruncatedAt
+	}
+
+	jEnd := a[j].EndTime
+	if a[j].Truncated() {
+		jEnd = a[j].TruncatedAt
+	}
+
+	if iEnd.Equal(jEnd) {
+		return a[i].StartTime.Before(a[j].StartTime)
+	}
+
+	return iEnd.Before(jEnd)
+}
+
+// Contains returns true if the shard group contains data for the timestamp.
 func (sgi *ShardGroupInfo) Contains(timestamp time.Time) bool {
 	return !sgi.StartTime.After(timestamp) && sgi.EndTime.After(timestamp)
 }
 
-// Overlaps return whether the shard group contains data for the time range between min and max
+// Overlaps returns whether the shard group contains data for the time range between min and max
 func (sgi *ShardGroupInfo) Overlaps(min, max time.Time) bool {
 	return !sgi.StartTime.After(max) && sgi.EndTime.After(min)
 }
@@ -1000,6 +1208,11 @@ func (sgi *ShardGroupInfo) Deleted() bool {
 	return !sgi.DeletedAt.IsZero()
 }
 
+// Truncated returns true if this ShardGroup has been truncated (no new writes).
+func (sgi *ShardGroupInfo) Truncated() bool {
+	return !sgi.TruncatedAt.IsZero()
+}
+
 // clone returns a deep copy of sgi.
 func (sgi ShardGroupInfo) clone() ShardGroupInfo {
 	other := sgi
@@ -1014,7 +1227,7 @@ func (sgi ShardGroupInfo) clone() ShardGroupInfo {
 	return other
 }
 
-// ShardFor returns the ShardInfo for a Point hash
+// ShardFor returns the ShardInfo for a Point hash.
 func (sgi *ShardGroupInfo) ShardFor(hash uint64) ShardInfo {
 	return sgi.Shards[hash%uint64(len(sgi.Shards))]
 }
@@ -1028,6 +1241,10 @@ func (sgi *ShardGroupInfo) marshal() *internal.ShardGroupInfo {
 		DeletedAt: proto.Int64(MarshalTime(sgi.DeletedAt)),
 	}
 
+	if !sgi.TruncatedAt.IsZero() {
+		pb.TruncatedAt = proto.Int64(MarshalTime(sgi.TruncatedAt))
+	}
+
 	pb.Shards = make([]*internal.ShardInfo, len(sgi.Shards))
 	for i := range sgi.Shards {
 		pb.Shards[i] = sgi.Shards[i].marshal()
@@ -1039,10 +1256,22 @@ func (sgi *ShardGroupInfo) marshal() *internal.ShardGroupInfo {
 // unmarshal deserializes from a protobuf representation.
 func (sgi *ShardGroupInfo) unmarshal(pb *internal.ShardGroupInfo) {
 	sgi.ID = pb.GetID()
-	sgi.StartTime = UnmarshalTime(pb.GetStartTime())
-	sgi.EndTime = UnmarshalTime(pb.GetEndTime())
+	if i := pb.GetStartTime(); i == 0 {
+		sgi.StartTime = time.Unix(0, 0).UTC()
+	} else {
+		sgi.StartTime = UnmarshalTime(i)
+	}
+	if i := pb.GetEndTime(); i == 0 {
+		sgi.EndTime = time.Unix(0, 0).UTC()
+	} else {
+		sgi.EndTime = UnmarshalTime(i)
+	}
 	sgi.DeletedAt = UnmarshalTime(pb.GetDeletedAt())
 
+	if pb != nil && pb.TruncatedAt != nil {
+		sgi.TruncatedAt = UnmarshalTime(pb.GetTruncatedAt())
+	}
+
 	if len(pb.GetShards()) > 0 {
 		sgi.Shards = make([]ShardInfo, len(pb.GetShards()))
 		for i, x := range pb.GetShards() {
@@ -1125,7 +1354,7 @@ func (si *ShardInfo) unmarshal(pb *internal.ShardInfo) {
 	}
 }
 
-// SubscriptionInfo hold the subscription information
+// SubscriptionInfo holds the subscription information.
 type SubscriptionInfo struct {
 	Name         string
 	Mode         string
@@ -1204,23 +1433,56 @@ func (cqi *ContinuousQueryInfo) unmarshal(pb *internal.ContinuousQueryInfo) {
 	cqi.Query = pb.GetQuery()
 }
 
+var _ query.Authorizer = (*UserInfo)(nil)
+
 // UserInfo represents metadata about a user in the system.
 type UserInfo struct {
-	Name       string
-	Hash       string
-	Admin      bool
+	// User's name.
+	Name string
+
+	// Hashed password.
+	Hash string
+
+	// Whether the user is an admin, i.e. allowed to do everything.
+	Admin bool
+
+	// Map of database name to granted privilege.
 	Privileges map[string]influxql.Privilege
 }
 
-// Authorize returns true if the user is authorized and false if not.
-func (ui *UserInfo) Authorize(privilege influxql.Privilege, database string) bool {
-	if ui.Admin {
+type User interface {
+	query.Authorizer
+	ID() string
+	IsAdmin() bool
+}
+
+func (u *UserInfo) ID() string {
+	return u.Name
+}
+
+func (u *UserInfo) IsAdmin() bool {
+	return u.Admin
+}
+
+// AuthorizeDatabase returns true if the user is authorized for the given privilege on the given database.
+func (ui *UserInfo) AuthorizeDatabase(privilege influxql.Privilege, database string) bool {
+	if ui.Admin || privilege == influxql.NoPrivileges {
 		return true
 	}
 	p, ok := ui.Privileges[database]
 	return ok && (p == privilege || p == influxql.AllPrivileges)
 }
 
+// AuthorizeSeriesRead is used to limit access per-series (enterprise only)
+func (u *UserInfo) AuthorizeSeriesRead(database string, measurement []byte, tags models.Tags) bool {
+	return true
+}
+
+// AuthorizeSeriesWrite is used to limit access per-series (enterprise only)
+func (u *UserInfo) AuthorizeSeriesWrite(database string, measurement []byte, tags models.Tags) bool {
+	return true
+}
+
 // clone returns a deep copy of si.
 func (ui UserInfo) clone() UserInfo {
 	other := ui
@@ -1265,18 +1527,21 @@ func (ui *UserInfo) unmarshal(pb *internal.UserInfo) {
 	}
 }
 
+// Lease represents a lease held on a resource.
 type Lease struct {
 	Name       string    `json:"name"`
 	Expiration time.Time `json:"expiration"`
 	Owner      uint64    `json:"owner"`
 }
 
+// Leases is a concurrency-safe collection of leases keyed by name.
 type Leases struct {
 	mu sync.Mutex
 	m  map[string]*Lease
 	d  time.Duration
 }
 
+// NewLeases returns a new instance of Leases.
 func NewLeases(d time.Duration) *Leases {
 	return &Leases{
 		m: make(map[string]*Lease),
@@ -1284,12 +1549,16 @@ func NewLeases(d time.Duration) *Leases {
 	}
 }
 
+// Acquire acquires a lease with the given name for the given nodeID.
+// If the lease doesn't exist or exists but is expired, a valid lease is returned.
+// If nodeID already owns the named and unexpired lease, the lease expiration is extended.
+// If a different node owns the lease, an error is returned.
 func (leases *Leases) Acquire(name string, nodeID uint64) (*Lease, error) {
 	leases.mu.Lock()
 	defer leases.mu.Unlock()
 
-	l, ok := leases.m[name]
-	if ok {
+	l := leases.m[name]
+	if l != nil {
 		if time.Now().After(l.Expiration) || l.Owner == nodeID {
 			l.Expiration = time.Now().Add(leases.d)
 			l.Owner = nodeID
@@ -1325,3 +1594,17 @@ func UnmarshalTime(v int64) time.Time {
 	}
 	return time.Unix(0, v).UTC()
 }
+
+// ValidName checks to see if the given name can would be valid for DB/RP name
+func ValidName(name string) bool {
+	for _, r := range name {
+		if !unicode.IsPrint(r) {
+			return false
+		}
+	}
+
+	return name != "" &&
+		name != "." &&
+		name != ".." &&
+		!strings.ContainsAny(name, `/\`)
+}
