@@ -1,10 +1,10 @@
 package tsm1
 
 import (
+	"bytes"
+	"context"
 	"fmt"
-	"io"
 	"io/ioutil"
-	"log"
 	"math"
 	"os"
 	"path/filepath"
@@ -16,60 +16,65 @@ import (
 	"time"
 
 	"github.com/influxdata/influxdb/models"
+	"github.com/influxdata/influxdb/pkg/metrics"
+	"github.com/influxdata/influxdb/query"
+	"github.com/uber-go/zap"
 )
 
+// TSMFile represents an on-disk TSM file.
 type TSMFile interface {
 	// Path returns the underlying file path for the TSMFile.  If the file
 	// has not be written or loaded from disk, the zero value is returned.
 	Path() string
 
-	// Read returns all the values in the block where time t resides
-	Read(key string, t int64) ([]Value, error)
+	// Read returns all the values in the block where time t resides.
+	Read(key []byte, t int64) ([]Value, error)
 
 	// ReadAt returns all the values in the block identified by entry.
 	ReadAt(entry *IndexEntry, values []Value) ([]Value, error)
-	ReadFloatBlockAt(entry *IndexEntry, tdec *TimeDecoder, vdec *FloatDecoder, values *[]FloatValue) ([]FloatValue, error)
-	ReadIntegerBlockAt(entry *IndexEntry, tdec *TimeDecoder, vdec *IntegerDecoder, values *[]IntegerValue) ([]IntegerValue, error)
-	ReadStringBlockAt(entry *IndexEntry, tdec *TimeDecoder, vdec *StringDecoder, values *[]StringValue) ([]StringValue, error)
-	ReadBooleanBlockAt(entry *IndexEntry, tdec *TimeDecoder, vdec *BooleanDecoder, values *[]BooleanValue) ([]BooleanValue, error)
+	ReadFloatBlockAt(entry *IndexEntry, values *[]FloatValue) ([]FloatValue, error)
+	ReadIntegerBlockAt(entry *IndexEntry, values *[]IntegerValue) ([]IntegerValue, error)
+	ReadUnsignedBlockAt(entry *IndexEntry, values *[]UnsignedValue) ([]UnsignedValue, error)
+	ReadStringBlockAt(entry *IndexEntry, values *[]StringValue) ([]StringValue, error)
+	ReadBooleanBlockAt(entry *IndexEntry, values *[]BooleanValue) ([]BooleanValue, error)
 
 	// Entries returns the index entries for all blocks for the given key.
-	Entries(key string) []IndexEntry
-	ReadEntries(key string, entries *[]IndexEntry)
+	Entries(key []byte) []IndexEntry
+	ReadEntries(key []byte, entries *[]IndexEntry) []IndexEntry
 
 	// Returns true if the TSMFile may contain a value with the specified
-	// key and time
-	ContainsValue(key string, t int64) bool
+	// key and time.
+	ContainsValue(key []byte, t int64) bool
 
 	// Contains returns true if the file contains any values for the given
 	// key.
-	Contains(key string) bool
+	Contains(key []byte) bool
 
 	// TimeRange returns the min and max time across all keys in the file.
 	TimeRange() (int64, int64)
 
 	// TombstoneRange returns ranges of time that are deleted for the given key.
-	TombstoneRange(key string) []TimeRange
+	TombstoneRange(key []byte) []TimeRange
 
 	// KeyRange returns the min and max keys in the file.
-	KeyRange() (string, string)
+	KeyRange() ([]byte, []byte)
 
-	// KeyCount returns the number of distict keys in the file.
+	// KeyCount returns the number of distinct keys in the file.
 	KeyCount() int
 
-	// KeyAt returns the key located at index position idx
-	KeyAt(idx int) (string, byte)
+	// KeyAt returns the key located at index position idx.
+	KeyAt(idx int) ([]byte, byte)
 
 	// Type returns the block type of the values stored for the key.  Returns one of
 	// BlockFloat64, BlockInt64, BlockBoolean, BlockString.  If key does not exist,
 	// an error is returned.
-	Type(key string) (byte, error)
+	Type(key []byte) (byte, error)
 
 	// Delete removes the keys from the set of keys available in this file.
-	Delete(keys []string) error
+	Delete(keys [][]byte) error
 
-	// DeleteRange removes the values for keys between min and max.
-	DeleteRange(keys []string, min, max int64) error
+	// DeleteRange removes the values for keys between timestamps min and max.
+	DeleteRange(keys [][]byte, min, max int64) error
 
 	// HasTombstones returns true if file contains values that have been deleted.
 	HasTombstones() bool
@@ -78,7 +83,7 @@ type TSMFile interface {
 	// written for this file.
 	TombstoneFiles() []FileStat
 
-	// Close the underlying file resources
+	// Close closes the underlying file resources.
 	Close() error
 
 	// Size returns the size of the file on disk in bytes.
@@ -88,24 +93,27 @@ type TSMFile interface {
 	// file name.  Index and Reader state are not re-initialized.
 	Rename(path string) error
 
-	// Remove deletes the file from the filesystem
+	// Remove deletes the file from the filesystem.
 	Remove() error
 
-	// Returns true if the file is currently in use by queries
+	// InUse returns true if the file is currently in use by queries.
 	InUse() bool
 
-	// Ref records that this file is actively in use
+	// Ref records that this file is actively in use.
 	Ref()
 
-	// Unref records that this file is no longer in user
+	// Unref records that this file is no longer in use.
 	Unref()
 
 	// Stats returns summary information about the TSM file.
 	Stats() FileStat
 
 	// BlockIterator returns an iterator pointing to the first block in the file and
-	// allows sequential iteration to each every block.
+	// allows sequential iteration to each and every block.
 	BlockIterator() *BlockIterator
+
+	// Free releases any resources held by the FileStore to free up system resources.
+	Free() error
 }
 
 // Statistics gathered by the FileStore.
@@ -114,18 +122,34 @@ const (
 	statFileStoreCount = "numFiles"
 )
 
+var (
+	floatBlocksDecodedCounter    = metrics.MustRegisterCounter("float_blocks_decoded", metrics.WithGroup(tsmGroup))
+	floatBlocksSizeCounter       = metrics.MustRegisterCounter("float_blocks_size_bytes", metrics.WithGroup(tsmGroup))
+	integerBlocksDecodedCounter  = metrics.MustRegisterCounter("integer_blocks_decoded", metrics.WithGroup(tsmGroup))
+	integerBlocksSizeCounter     = metrics.MustRegisterCounter("integer_blocks_size_bytes", metrics.WithGroup(tsmGroup))
+	unsignedBlocksDecodedCounter = metrics.MustRegisterCounter("unsigned_blocks_decoded", metrics.WithGroup(tsmGroup))
+	unsignedBlocksSizeCounter    = metrics.MustRegisterCounter("unsigned_blocks_size_bytes", metrics.WithGroup(tsmGroup))
+	stringBlocksDecodedCounter   = metrics.MustRegisterCounter("string_blocks_decoded", metrics.WithGroup(tsmGroup))
+	stringBlocksSizeCounter      = metrics.MustRegisterCounter("string_blocks_size_bytes", metrics.WithGroup(tsmGroup))
+	booleanBlocksDecodedCounter  = metrics.MustRegisterCounter("boolean_blocks_decoded", metrics.WithGroup(tsmGroup))
+	booleanBlocksSizeCounter     = metrics.MustRegisterCounter("boolean_blocks_size_bytes", metrics.WithGroup(tsmGroup))
+)
+
+// FileStore is an abstraction around multiple TSM files.
 type FileStore struct {
 	mu           sync.RWMutex
 	lastModified time.Time
+	// Most recently known file stats. If nil then stats will need to be
+	// recalculated
+	lastFileStats []FileStat
 
 	currentGeneration int
 	dir               string
 
 	files []TSMFile
 
-	logger       *log.Logger // Logger to be used for important messages
-	traceLogger  *log.Logger // Logger to be used when trace-logging is on.
-	logOutput    io.Writer   // Writer to be logger and traceLogger if active.
+	logger       zap.Logger // Logger to be used for important messages
+	traceLogger  zap.Logger // Logger to be used when trace-logging is on.
 	traceLogging bool
 
 	stats  *FileStoreStatistics
@@ -134,64 +158,65 @@ type FileStore struct {
 	currentTempDirID int
 }
 
+// FileStat holds information about a TSM file on disk.
 type FileStat struct {
 	Path             string
 	HasTombstone     bool
 	Size             uint32
 	LastModified     int64
 	MinTime, MaxTime int64
-	MinKey, MaxKey   string
+	MinKey, MaxKey   []byte
 }
 
+// OverlapsTimeRange returns true if the time range of the file intersect min and max.
 func (f FileStat) OverlapsTimeRange(min, max int64) bool {
 	return f.MinTime <= max && f.MaxTime >= min
 }
 
-func (f FileStat) OverlapsKeyRange(min, max string) bool {
-	return min != "" && max != "" && f.MinKey <= max && f.MaxKey >= min
+// OverlapsKeyRange returns true if the min and max keys of the file overlap the arguments min and max.
+func (f FileStat) OverlapsKeyRange(min, max []byte) bool {
+	return len(min) != 0 && len(max) != 0 && bytes.Compare(f.MinKey, max) <= 0 && bytes.Compare(f.MaxKey, min) >= 0
 }
 
-func (f FileStat) ContainsKey(key string) bool {
-	return f.MinKey >= key || key <= f.MaxKey
+// ContainsKey returns true if the min and max keys of the file overlap the arguments min and max.
+func (f FileStat) ContainsKey(key []byte) bool {
+	return bytes.Compare(f.MinKey, key) >= 0 || bytes.Compare(key, f.MaxKey) <= 0
 }
 
+// NewFileStore returns a new instance of FileStore based on the given directory.
 func NewFileStore(dir string) *FileStore {
-	logger := log.New(os.Stderr, "[filestore] ", log.LstdFlags)
-	return &FileStore{
+	logger := zap.New(zap.NullEncoder())
+	fs := &FileStore{
 		dir:          dir,
-		lastModified: time.Now(),
+		lastModified: time.Time{},
 		logger:       logger,
-		traceLogger:  log.New(ioutil.Discard, "[filestore] ", log.LstdFlags),
-		logOutput:    os.Stderr,
+		traceLogger:  logger,
 		stats:        &FileStoreStatistics{},
 		purger: &purger{
 			files:  map[string]TSMFile{},
 			logger: logger,
 		},
 	}
+	fs.purger.fileStore = fs
+	return fs
 }
 
 // enableTraceLogging must be called before the FileStore is opened.
 func (f *FileStore) enableTraceLogging(enabled bool) {
 	f.traceLogging = enabled
 	if enabled {
-		f.traceLogger.SetOutput(f.logOutput)
+		f.traceLogger = f.logger
 	}
 }
 
-// SetLogOutput sets the logger used for all messages. It is safe for concurrent
-// use.
-func (f *FileStore) SetLogOutput(w io.Writer) {
-	f.logger.SetOutput(w)
+// WithLogger sets the logger on the file store.
+func (f *FileStore) WithLogger(log zap.Logger) {
+	f.logger = log.With(zap.String("service", "filestore"))
+	f.purger.logger = f.logger
 
-	// Set the trace logger's output only if trace logging is enabled.
 	if f.traceLogging {
-		f.traceLogger.SetOutput(w)
+		f.traceLogger = f.logger
 	}
-
-	f.mu.Lock()
-	f.logOutput = w
-	f.mu.Unlock()
 }
 
 // FileStoreStatistics keeps statistics about the file store.
@@ -212,28 +237,41 @@ func (f *FileStore) Statistics(tags map[string]string) []models.Statistic {
 	}}
 }
 
-// Returns the number of TSM files currently loaded
+// Count returns the number of TSM files currently loaded.
 func (f *FileStore) Count() int {
 	f.mu.RLock()
 	defer f.mu.RUnlock()
 	return len(f.files)
 }
 
-// Files returns TSM files currently loaded.
+// Files returns the slice of TSM files currently loaded.
 func (f *FileStore) Files() []TSMFile {
 	f.mu.RLock()
 	defer f.mu.RUnlock()
 	return f.files
 }
 
-// CurrentGeneration returns the current generation of the TSM files
+// Free releases any resources held by the FileStore.  The resources will be re-acquired
+// if necessary if they are needed after freeing them.
+func (f *FileStore) Free() error {
+	f.mu.RLock()
+	defer f.mu.RUnlock()
+	for _, f := range f.files {
+		if err := f.Free(); err != nil {
+			return err
+		}
+	}
+	return nil
+}
+
+// CurrentGeneration returns the current generation of the TSM files.
 func (f *FileStore) CurrentGeneration() int {
 	f.mu.RLock()
 	defer f.mu.RUnlock()
 	return f.currentGeneration
 }
 
-// NextGeneration returns the max file ID + 1
+// NextGeneration increments the max file ID and returns the new value.
 func (f *FileStore) NextGeneration() int {
 	f.mu.Lock()
 	defer f.mu.Unlock()
@@ -241,78 +279,59 @@ func (f *FileStore) NextGeneration() int {
 	return f.currentGeneration
 }
 
-func (f *FileStore) Add(files ...TSMFile) {
-	f.mu.Lock()
-	defer f.mu.Unlock()
-	for _, file := range files {
-		atomic.AddInt64(&f.stats.DiskBytes, int64(file.Size()))
-	}
-	f.files = append(f.files, files...)
-	sort.Sort(tsmReaders(f.files))
-	atomic.StoreInt64(&f.stats.FileCount, int64(len(f.files)))
-}
-
-// Remove removes the files with matching paths from the set of active files.  It does
-// not remove the paths from disk.
-func (f *FileStore) Remove(paths ...string) {
-	f.mu.Lock()
-	defer f.mu.Unlock()
-	var active []TSMFile
-	for _, file := range f.files {
-		keep := true
-		for _, remove := range paths {
-			if remove == file.Path() {
-				keep = false
-				break
-			}
-		}
-
-		if keep {
-			active = append(active, file)
-		} else {
-			// Removing the file, remove the file size from the total file store bytes
-			atomic.AddInt64(&f.stats.DiskBytes, -int64(file.Size()))
-		}
-	}
-	f.files = active
-	sort.Sort(tsmReaders(f.files))
-	atomic.StoreInt64(&f.stats.FileCount, int64(len(f.files)))
-}
-
 // WalkKeys calls fn for every key in every TSM file known to the FileStore.  If the key
 // exists in multiple files, it will be invoked for each file.
-func (f *FileStore) WalkKeys(fn func(key string, typ byte) error) error {
+func (f *FileStore) WalkKeys(fn func(key []byte, typ byte) error) error {
 	f.mu.RLock()
-	defer f.mu.RUnlock()
+	if len(f.files) == 0 {
+		f.mu.RUnlock()
+		return nil
+	}
 
+	readers := make([]chan seriesKey, 0, len(f.files))
 	for _, f := range f.files {
-		for i := 0; i < f.KeyCount(); i++ {
-			key, typ := f.KeyAt(i)
-			if err := fn(key, typ); err != nil {
-				return err
+		ch := make(chan seriesKey, 1)
+		readers = append(readers, ch)
+
+		go func(c chan seriesKey, r TSMFile) {
+			n := r.KeyCount()
+			for i := 0; i < n; i++ {
+				key, typ := r.KeyAt(i)
+				c <- seriesKey{key, typ}
 			}
+			close(ch)
+		}(ch, f)
+	}
+	f.mu.RUnlock()
+
+	merged := merge(readers...)
+	for v := range merged {
+		if err := fn(v.key, v.typ); err != nil {
+			return err
 		}
 	}
+
 	return nil
 }
 
-// Keys returns all keys and types for all files
+// Keys returns all keys and types for all files in the file store.
 func (f *FileStore) Keys() map[string]byte {
 	f.mu.RLock()
 	defer f.mu.RUnlock()
 
 	uniqueKeys := map[string]byte{}
-	for _, f := range f.files {
-		for i := 0; i < f.KeyCount(); i++ {
-			key, typ := f.KeyAt(i)
-			uniqueKeys[key] = typ
-		}
+	if err := f.WalkKeys(func(key []byte, typ byte) error {
+		uniqueKeys[string(key)] = typ
+		return nil
+	}); err != nil {
+		return nil
 	}
 
 	return uniqueKeys
 }
 
-func (f *FileStore) Type(key string) (byte, error) {
+// Type returns the type of values store at the block for key.
+func (f *FileStore) Type(key []byte) (byte, error) {
 	f.mu.RLock()
 	defer f.mu.RUnlock()
 
@@ -324,21 +343,27 @@ func (f *FileStore) Type(key string) (byte, error) {
 	return 0, fmt.Errorf("unknown type for %v", key)
 }
 
-func (f *FileStore) Delete(keys []string) error {
+// Delete removes the keys from the set of keys available in this file.
+func (f *FileStore) Delete(keys [][]byte) error {
 	return f.DeleteRange(keys, math.MinInt64, math.MaxInt64)
 }
 
-// DeleteRange removes the values for keys between min and max.
-func (f *FileStore) DeleteRange(keys []string, min, max int64) error {
-	f.mu.Lock()
-	f.lastModified = time.Now()
-	f.mu.Unlock()
-
-	return f.walkFiles(func(tsm TSMFile) error {
+// DeleteRange removes the values for keys between timestamps min and max.
+func (f *FileStore) DeleteRange(keys [][]byte, min, max int64) error {
+	if err := f.walkFiles(func(tsm TSMFile) error {
 		return tsm.DeleteRange(keys, min, max)
-	})
+	}); err != nil {
+		return err
+	}
+
+	f.mu.Lock()
+	f.lastModified = time.Now().UTC()
+	f.lastFileStats = nil
+	f.mu.Unlock()
+	return nil
 }
 
+// Open loads all the TSM files in the configured directory.
 func (f *FileStore) Open() error {
 	f.mu.Lock()
 	defer f.mu.Unlock()
@@ -394,15 +419,10 @@ func (f *FileStore) Open() error {
 			return fmt.Errorf("error opening file %s: %v", fn, err)
 		}
 
-		// Accumulate file store size stat
-		if fi, err := file.Stat(); err == nil {
-			atomic.AddInt64(&f.stats.DiskBytes, fi.Size())
-		}
-
 		go func(idx int, file *os.File) {
 			start := time.Now()
 			df, err := NewTSMReader(file)
-			f.logger.Printf("%s (#%d) opened in %v", file.Name(), idx, time.Now().Sub(start))
+			f.logger.Info(fmt.Sprintf("%s (#%d) opened in %v", file.Name(), idx, time.Since(start)))
 
 			if err != nil {
 				readerC <- &res{r: df, err: fmt.Errorf("error opening memory map for file %s: %v", file.Name(), err)}
@@ -412,6 +432,7 @@ func (f *FileStore) Open() error {
 		}(i, file)
 	}
 
+	var lm int64
 	for range files {
 		res := <-readerC
 		if res.err != nil {
@@ -419,7 +440,19 @@ func (f *FileStore) Open() error {
 			return res.err
 		}
 		f.files = append(f.files, res.r)
+		// Accumulate file store size stats
+		atomic.AddInt64(&f.stats.DiskBytes, int64(res.r.Size()))
+		for _, ts := range res.r.TombstoneFiles() {
+			atomic.AddInt64(&f.stats.DiskBytes, int64(ts.Size))
+		}
+
+		// Re-initialize the lastModified time for the file store
+		if res.r.LastModified() > lm {
+			lm = res.r.LastModified()
+		}
+
 	}
+	f.lastModified = time.Unix(0, lm)
 	close(readerC)
 
 	sort.Sort(tsmReaders(f.files))
@@ -427,20 +460,28 @@ func (f *FileStore) Open() error {
 	return nil
 }
 
+// Close closes the file store.
 func (f *FileStore) Close() error {
 	f.mu.Lock()
 	defer f.mu.Unlock()
 
-	for _, f := range f.files {
-		f.Close()
+	for _, file := range f.files {
+		file.Close()
 	}
 
+	f.lastFileStats = nil
 	f.files = nil
 	atomic.StoreInt64(&f.stats.FileCount, 0)
 	return nil
 }
 
-func (f *FileStore) Read(key string, t int64) ([]Value, error) {
+func (f *FileStore) DiskSizeBytes() int64 {
+	return atomic.LoadInt64(&f.stats.DiskBytes)
+}
+
+// Read returns the slice of values for the given key and the given timestamp,
+// if any file matches those constraints.
+func (f *FileStore) Read(key []byte, t int64) ([]Value, error) {
 	f.mu.RLock()
 	defer f.mu.RUnlock()
 
@@ -463,38 +504,83 @@ func (f *FileStore) Read(key string, t int64) ([]Value, error) {
 	return nil, nil
 }
 
-func (f *FileStore) KeyCursor(key string, t int64, ascending bool) *KeyCursor {
+func (f *FileStore) Cost(key []byte, min, max int64) query.IteratorCost {
 	f.mu.RLock()
 	defer f.mu.RUnlock()
-	return newKeyCursor(f, key, t, ascending)
+	return f.cost(key, min, max)
 }
 
+// Reader returns a TSMReader for path if one is currently managed by the FileStore.
+// Otherwise it returns nil.
+func (f *FileStore) TSMReader(path string) *TSMReader {
+	f.mu.RLock()
+	defer f.mu.RUnlock()
+	for _, r := range f.files {
+		if r.Path() == path {
+			return r.(*TSMReader)
+		}
+	}
+	return nil
+}
+
+// KeyCursor returns a KeyCursor for key and t across the files in the FileStore.
+func (f *FileStore) KeyCursor(ctx context.Context, key []byte, t int64, ascending bool) *KeyCursor {
+	f.mu.RLock()
+	defer f.mu.RUnlock()
+	return newKeyCursor(ctx, f, key, t, ascending)
+}
+
+// Stats returns the stats of the underlying files, preferring the cached version if it is still valid.
 func (f *FileStore) Stats() []FileStat {
 	f.mu.RLock()
-	defer f.mu.RUnlock()
-	stats := make([]FileStat, len(f.files))
-	for i, fd := range f.files {
-		stats[i] = fd.Stats()
+	if len(f.lastFileStats) > 0 {
+		defer f.mu.RUnlock()
+		return f.lastFileStats
 	}
+	f.mu.RUnlock()
 
-	return stats
-}
-
-func (f *FileStore) Replace(oldFiles, newFiles []string) error {
+	// The file stats cache is invalid due to changes to files. Need to
+	// recalculate.
 	f.mu.Lock()
 	defer f.mu.Unlock()
 
-	f.lastModified = time.Now()
-
-	// Copy the current set of active files while we rename
-	// and load the new files.  We copy the pointers here to minimize
-	// the time that locks are held as well as to ensure that the replacement
-	// is atomic.©
-	var updated []TSMFile
-	for _, t := range f.files {
-		updated = append(updated, t)
+	if len(f.lastFileStats) > 0 {
+		return f.lastFileStats
 	}
 
+	// If lastFileStats's capacity is far away from the number of entries
+	// we need to add, then we'll reallocate.
+	if cap(f.lastFileStats) < len(f.files)/2 {
+		f.lastFileStats = make([]FileStat, 0, len(f.files))
+	}
+
+	for _, fd := range f.files {
+		f.lastFileStats = append(f.lastFileStats, fd.Stats())
+	}
+	return f.lastFileStats
+}
+
+// ReplaceWithCallback replaces oldFiles with newFiles and calls updatedFn with the files to be added the FileStore.
+func (f *FileStore) ReplaceWithCallback(oldFiles, newFiles []string, updatedFn func(r []TSMFile)) error {
+	return f.replace(oldFiles, newFiles, updatedFn)
+}
+
+// Replace replaces oldFiles with newFiles.
+func (f *FileStore) Replace(oldFiles, newFiles []string) error {
+	return f.replace(oldFiles, newFiles, nil)
+}
+
+func (f *FileStore) replace(oldFiles, newFiles []string, updatedFn func(r []TSMFile)) error {
+	if len(oldFiles) == 0 && len(newFiles) == 0 {
+		return nil
+	}
+
+	f.mu.RLock()
+	maxTime := f.lastModified
+	f.mu.RUnlock()
+
+	updated := make([]TSMFile, 0, len(newFiles))
+
 	// Rename all the new files to make them live on restart
 	for _, file := range newFiles {
 		var newName = file
@@ -511,6 +597,13 @@ func (f *FileStore) Replace(oldFiles, newFiles []string) error {
 			return err
 		}
 
+		// Keep track of the new mod time
+		if stat, err := fd.Stat(); err == nil {
+			if stat.ModTime().UTC().After(maxTime) {
+				maxTime = stat.ModTime().UTC()
+			}
+		}
+
 		tsm, err := NewTSMReader(fd)
 		if err != nil {
 			return err
@@ -518,6 +611,20 @@ func (f *FileStore) Replace(oldFiles, newFiles []string) error {
 		updated = append(updated, tsm)
 	}
 
+	if updatedFn != nil {
+		updatedFn(updated)
+	}
+
+	f.mu.Lock()
+	defer f.mu.Unlock()
+
+	// Copy the current set of active files while we rename
+	// and load the new files.  We copy the pointers here to minimize
+	// the time that locks are held as well as to ensure that the replacement
+	// is atomic.©
+
+	updated = append(updated, f.files...)
+
 	// We need to prune our set of active files now
 	var active, inuse []TSMFile
 	for _, file := range updated {
@@ -579,6 +686,16 @@ func (f *FileStore) Replace(oldFiles, newFiles []string) error {
 	// Tell the purger about our in-use files we need to remove
 	f.purger.add(inuse)
 
+	// If times didn't change (which can happen since file mod times are second level),
+	// then add a ns to the time to ensure that lastModified changes since files on disk
+	// actually did change
+	if maxTime.Equal(f.lastModified) || maxTime.Before(f.lastModified) {
+		maxTime = f.lastModified.UTC().Add(1)
+	}
+
+	f.lastModified = maxTime.UTC()
+
+	f.lastFileStats = nil
 	f.files = active
 	sort.Sort(tsmReaders(f.files))
 	atomic.StoreInt64(&f.stats.FileCount, int64(len(f.files)))
@@ -587,6 +704,10 @@ func (f *FileStore) Replace(oldFiles, newFiles []string) error {
 	var totalSize int64
 	for _, file := range f.files {
 		totalSize += int64(file.Size())
+		for _, ts := range file.TombstoneFiles() {
+			totalSize += int64(ts.Size)
+		}
+
 	}
 	atomic.StoreInt64(&f.stats.DiskBytes, totalSize)
 
@@ -594,7 +715,7 @@ func (f *FileStore) Replace(oldFiles, newFiles []string) error {
 }
 
 // LastModified returns the last time the file store was updated with new
-// TSM files or a delete
+// TSM files or a delete.
 func (f *FileStore) LastModified() time.Time {
 	f.mu.RLock()
 	defer f.mu.RUnlock()
@@ -622,22 +743,25 @@ func (f *FileStore) BlockCount(path string, idx int) int {
 					return 0
 				}
 			}
-			_, _, _, _, block, _ := iter.Read()
+			_, _, _, _, _, block, _ := iter.Read()
 			return BlockCount(block)
 		}
 	}
 	return 0
 }
 
-// walkFiles calls fn for every files in filestore in parallel
+// walkFiles calls fn for each file in filestore in parallel.
 func (f *FileStore) walkFiles(fn func(f TSMFile) error) error {
+	// Copy the current TSM files to prevent a slow walker from
+	// blocking other operations.
 	f.mu.RLock()
-	defer f.mu.RUnlock()
+	files := make([]TSMFile, len(f.files))
+	copy(files, f.files)
+	f.mu.RUnlock()
 
 	// struct to hold the result of opening each reader in a goroutine
-
-	errC := make(chan error, len(f.files))
-	for _, f := range f.files {
+	errC := make(chan error, len(files))
+	for _, f := range files {
 		go func(tsm TSMFile) {
 			if err := fn(tsm); err != nil {
 				errC <- fmt.Errorf("file %s: %s", tsm.Path(), err)
@@ -657,22 +781,56 @@ func (f *FileStore) walkFiles(fn func(f TSMFile) error) error {
 	return nil
 }
 
+// We need to determine the possible files that may be accessed by this query given
+// the time range.
+func (f *FileStore) cost(key []byte, min, max int64) query.IteratorCost {
+	var cache []IndexEntry
+	cost := query.IteratorCost{}
+	for _, fd := range f.files {
+		minTime, maxTime := fd.TimeRange()
+		if !(maxTime > min && minTime < max) {
+			continue
+		}
+		skipped := true
+		tombstones := fd.TombstoneRange(key)
+
+		entries := fd.ReadEntries(key, &cache)
+	ENTRIES:
+		for i := 0; i < len(entries); i++ {
+			ie := entries[i]
+
+			if !(ie.MaxTime > min && ie.MinTime < max) {
+				continue
+			}
+
+			// Skip any blocks only contain values that are tombstoned.
+			for _, t := range tombstones {
+				if t.Min <= ie.MinTime && t.Max >= ie.MaxTime {
+					continue ENTRIES
+				}
+			}
+
+			cost.BlocksRead++
+			cost.BlockSize += int64(ie.Size)
+			skipped = false
+		}
+
+		if !skipped {
+			cost.NumFiles++
+		}
+	}
+	return cost
+}
+
 // locations returns the files and index blocks for a key and time.  ascending indicates
 // whether the key will be scan in ascending time order or descenging time order.
 // This function assumes the read-lock has been taken.
-func (f *FileStore) locations(key string, t int64, ascending bool) []*location {
-	var locations []*location
-
-	filesSnapshot := make([]TSMFile, len(f.files))
-	for i := range f.files {
-		filesSnapshot[i] = f.files[i]
-	}
-
-	var entries []IndexEntry
-	for _, fd := range filesSnapshot {
+func (f *FileStore) locations(key []byte, t int64, ascending bool) []*location {
+	var cache []IndexEntry
+	locations := make([]*location, 0, len(f.files))
+	for _, fd := range f.files {
 		minTime, maxTime := fd.TimeRange()
 
-		tombstones := fd.TombstoneRange(key)
 		// If we ascending and the max time of the file is before where we want to start
 		// skip it.
 		if ascending && maxTime < t {
@@ -682,24 +840,22 @@ func (f *FileStore) locations(key string, t int64, ascending bool) []*location {
 		} else if !ascending && minTime > t {
 			continue
 		}
+		tombstones := fd.TombstoneRange(key)
 
 		// This file could potential contain points we are looking for so find the blocks for
 		// the given key.
-		fd.ReadEntries(key, &entries)
-		for _, ie := range entries {
+		entries := fd.ReadEntries(key, &cache)
+	LOOP:
+		for i := 0; i < len(entries); i++ {
+			ie := entries[i]
 
 			// Skip any blocks only contain values that are tombstoned.
-			var skip bool
 			for _, t := range tombstones {
 				if t.Min <= ie.MinTime && t.Max >= ie.MaxTime {
-					skip = true
-					break
+					continue LOOP
 				}
 			}
 
-			if skip {
-				continue
-			}
 			// If we ascending and the max time of a block is before where we are looking, skip
 			// it since the data is out of our range
 			if ascending && ie.MaxTime < t {
@@ -733,10 +889,10 @@ func (f *FileStore) locations(key string, t int64, ascending bool) []*location {
 	return locations
 }
 
-// CreateSnapshot will create hardlinks for all tsm and tombstone files
-// in the path provided
+// CreateSnapshot creates hardlinks for all tsm and tombstone files
+// in the path provided.
 func (f *FileStore) CreateSnapshot() (string, error) {
-	f.traceLogger.Printf("Creating snapshot in %s", f.dir)
+	f.traceLogger.Info(fmt.Sprintf("Creating snapshot in %s", f.dir))
 	files := f.Files()
 
 	f.mu.Lock()
@@ -780,20 +936,27 @@ func ParseTSMFileName(name string) (int, int, error) {
 
 	id := base[:idx]
 
-	parts := strings.Split(id, "-")
-	if len(parts) != 2 {
+	idx = strings.Index(id, "-")
+	if idx == -1 {
 		return 0, 0, fmt.Errorf("file %s is named incorrectly", name)
 	}
 
-	generation, err := strconv.ParseUint(parts[0], 10, 32)
-	sequence, err := strconv.ParseUint(parts[1], 10, 32)
+	generation, err := strconv.ParseUint(id[:idx], 10, 32)
+	if err != nil {
+		return 0, 0, fmt.Errorf("file %s is named incorrectly", name)
+	}
 
-	return int(generation), int(sequence), err
+	sequence, err := strconv.ParseUint(id[idx+1:], 10, 32)
+	if err != nil {
+		return 0, 0, fmt.Errorf("file %s is named incorrectly", name)
+	}
+
+	return int(generation), int(sequence), nil
 }
 
+// KeyCursor allows iteration through keys in a set of files within a FileStore.
 type KeyCursor struct {
-	key string
-	fs  *FileStore
+	key []byte
 
 	// seeks is all the file locations that we need to return during iteration.
 	seeks []*location
@@ -804,19 +967,13 @@ type KeyCursor struct {
 	current []*location
 	buf     []Value
 
+	ctx context.Context
+	col *metrics.Group
+
 	// pos is the index within seeks.  Based on ascending, it will increment or
 	// decrement through the size of seeks slice.
 	pos       int
 	ascending bool
-
-	// duplicates is a hint that there are overlapping blocks for this key in
-	// multiple files (e.g. points have been overwritten but not fully compacted)
-	// If this is true, we need to scan the duplicate blocks and dedup the points
-	// as query time until they are compacted.
-	duplicates bool
-
-	// The distinct set of TSM files references by the cursor
-	refs map[string]TSMFile
 }
 
 type location struct {
@@ -866,15 +1023,14 @@ func (a ascLocations) Less(i, j int) bool {
 
 // newKeyCursor returns a new instance of KeyCursor.
 // This function assumes the read-lock has been taken.
-func newKeyCursor(fs *FileStore, key string, t int64, ascending bool) *KeyCursor {
+func newKeyCursor(ctx context.Context, fs *FileStore, key []byte, t int64, ascending bool) *KeyCursor {
 	c := &KeyCursor{
 		key:       key,
-		fs:        fs,
 		seeks:     fs.locations(key, t, ascending),
+		ctx:       ctx,
+		col:       metrics.GroupFromContext(ctx),
 		ascending: ascending,
-		refs:      map[string]TSMFile{},
 	}
-	c.duplicates = c.hasOverlappingBlocks()
 
 	if ascending {
 		sort.Sort(ascLocations(c.seeks))
@@ -884,10 +1040,7 @@ func newKeyCursor(fs *FileStore, key string, t int64, ascending bool) *KeyCursor
 
 	// Determine the distinct set of TSM files in use and mark then as in-use
 	for _, f := range c.seeks {
-		if _, ok := c.refs[f.r.Path()]; !ok {
-			f.r.Ref()
-			c.refs[f.r.Path()] = f.r
-		}
+		f.r.Ref()
 	}
 
 	c.seek(t)
@@ -897,13 +1050,12 @@ func newKeyCursor(fs *FileStore, key string, t int64, ascending bool) *KeyCursor
 // Close removes all references on the cursor.
 func (c *KeyCursor) Close() {
 	// Remove all of our in-use references since we're done
-	for _, f := range c.refs {
-		f.Unref()
+	for _, f := range c.seeks {
+		f.r.Unref()
 	}
 
 	c.buf = nil
 	c.seeks = nil
-	c.fs = nil
 	c.current = nil
 }
 
@@ -947,12 +1099,6 @@ func (c *KeyCursor) seekAscending(t int64) {
 			}
 
 			c.current = append(c.current, e)
-
-			// Exit if we don't have duplicates.
-			// Otherwise, keep looking for additional blocks containing this point.
-			if !c.duplicates {
-				return
-			}
 		}
 	}
 }
@@ -966,12 +1112,6 @@ func (c *KeyCursor) seekDescending(t int64) {
 				c.pos = i
 			}
 			c.current = append(c.current, e)
-
-			// Exit if we don't have duplicates.
-			// Otherwise, keep looking for additional blocks containing this point.
-			if !c.duplicates {
-				return
-			}
 		}
 	}
 }
@@ -1006,17 +1146,12 @@ func (c *KeyCursor) nextAscending() {
 
 	// Append the first matching block
 	if len(c.current) == 0 {
-		c.current = append(c.current, &location{})
+		c.current = append(c.current, nil)
 	} else {
 		c.current = c.current[:1]
 	}
 	c.current[0] = c.seeks[c.pos]
 
-	// We're done if there are no overlapping blocks.
-	if !c.duplicates {
-		return
-	}
-
 	// If we have ovelapping blocks, append all their values so we can dedup
 	for i := c.pos + 1; i < len(c.seeks); i++ {
 		if c.seeks[i].read() {
@@ -1045,11 +1180,6 @@ func (c *KeyCursor) nextDescending() {
 	}
 	c.current[0] = c.seeks[c.pos]
 
-	// We're done if there are no overlapping blocks.
-	if !c.duplicates {
-		return
-	}
-
 	// If we have ovelapping blocks, append all their values so we can dedup
 	for i := c.pos; i >= 0; i-- {
 		if c.seeks[i].read() {
@@ -1073,6 +1203,13 @@ func (c *KeyCursor) filterIntegerValues(tombstones []TimeRange, values IntegerVa
 	return values
 }
 
+func (c *KeyCursor) filterUnsignedValues(tombstones []TimeRange, values UnsignedValues) UnsignedValues {
+	for _, t := range tombstones {
+		values = values.Exclude(t.Min, t.Max)
+	}
+	return values
+}
+
 func (c *KeyCursor) filterStringValues(tombstones []TimeRange, values StringValues) StringValues {
 	for _, t := range tombstones {
 		values = values.Exclude(t.Min, t.Max)
@@ -1088,11 +1225,12 @@ func (c *KeyCursor) filterBooleanValues(tombstones []TimeRange, values BooleanVa
 }
 
 type purger struct {
-	mu      sync.RWMutex
-	files   map[string]TSMFile
-	running bool
+	mu        sync.RWMutex
+	fileStore *FileStore
+	files     map[string]TSMFile
+	running   bool
 
-	logger *log.Logger
+	logger zap.Logger
 }
 
 func (p *purger) add(files []TSMFile) {
@@ -1119,12 +1257,12 @@ func (p *purger) purge() {
 			for k, v := range p.files {
 				if !v.InUse() {
 					if err := v.Close(); err != nil {
-						p.logger.Printf("purge: close file: %v", err)
+						p.logger.Info(fmt.Sprintf("purge: close file: %v", err))
 						continue
 					}
 
 					if err := v.Remove(); err != nil {
-						p.logger.Printf("purge: remove file: %v", err)
+						p.logger.Info(fmt.Sprintf("purge: remove file: %v", err))
 						continue
 					}
 					delete(p.files, k)
@@ -1148,3 +1286,103 @@ type tsmReaders []TSMFile
 func (a tsmReaders) Len() int           { return len(a) }
 func (a tsmReaders) Less(i, j int) bool { return a[i].Path() < a[j].Path() }
 func (a tsmReaders) Swap(i, j int)      { a[i], a[j] = a[j], a[i] }
+
+type stream struct {
+	c chan seriesKey
+	v seriesKey
+}
+
+type seriesKey struct {
+	key []byte
+	typ byte
+}
+
+// merge merges multiple channels in parallel by recursively splitting the channels
+// until a simple two-way merge can be performed.
+func merge(c ...chan seriesKey) chan seriesKey {
+	if len(c) == 0 {
+		m := make(chan seriesKey)
+		close(m)
+		return m
+	}
+
+	// Just one, drain it
+	if len(c) == 1 {
+		m := make(chan seriesKey)
+		go func() {
+			if c[0] != nil {
+				for v := range c[0] {
+					m <- v
+				}
+			}
+			close(m)
+		}()
+		return m
+	}
+
+	// More than two, split them up recursively
+	if len(c) > 2 {
+		a := merge(c[:len(c)/2]...)
+		b := merge(c[len(c)/2:]...)
+		return merge(a, b)
+	}
+
+	// Merge the two streams and drop duplicates between then
+	m := make(chan seriesKey, 1)
+	a, b := c[0], c[1]
+	go func() {
+		// buffer a and b values
+		var av, bv seriesKey
+		if a != nil {
+			av = <-a
+		}
+		if b != nil {
+			bv = <-b
+		}
+		for {
+			if len(av.key) == 0 && len(bv.key) == 0 {
+				break
+			}
+
+			if len(av.key) == 0 {
+				m <- bv
+				break
+			}
+
+			if len(bv.key) == 0 {
+				m <- av
+				break
+			}
+
+			cmp := bytes.Compare(av.key, bv.key)
+			if cmp < 0 {
+				// Send a's value, and re-prime a buffer
+				m <- av
+				av = <-a
+			} else if cmp == 0 {
+				// Send a's value, and re-prime a and b buffers
+				m <- av
+				av = <-a
+				bv = <-b
+			} else {
+				// Send b's value, and re-prime b buffer
+				m <- bv
+				bv = <-b
+			}
+		}
+
+		if a != nil {
+			for av := range a {
+				m <- av
+			}
+		}
+
+		if b != nil {
+			for bv := range b {
+				m <- bv
+			}
+		}
+		close(m)
+	}()
+	return m
+}
