@@ -2,6 +2,7 @@ package tsm1
 
 import (
 	"bufio"
+	"compress/gzip"
 	"encoding/binary"
 	"io"
 	"io/ioutil"
@@ -13,8 +14,9 @@ import (
 )
 
 const (
-	v2header     = 0x1502
-	v2headerSize = 4
+	headerSize = 4
+	v2header   = 0x1502
+	v3header   = 0x1503
 )
 
 // Tombstoner records tombstones when entries are deleted.
@@ -35,7 +37,7 @@ type Tombstoner struct {
 // Tombstone represents an individual deletion.
 type Tombstone struct {
 	// Key is the tombstoned series key.
-	Key string
+	Key []byte
 
 	// Min and Max are the min and max unix nanosecond time ranges of Key that are deleted.  If
 	// the full range is deleted, both values are -1.
@@ -43,12 +45,12 @@ type Tombstone struct {
 }
 
 // Add adds the all keys, across all timestamps, to the tombstone.
-func (t *Tombstoner) Add(keys []string) error {
+func (t *Tombstoner) Add(keys [][]byte) error {
 	return t.AddRange(keys, math.MinInt64, math.MaxInt64)
 }
 
 // AddRange adds all keys to the tombstone specifying only the data between min and max to be removed.
-func (t *Tombstoner) AddRange(keys []string, min, max int64) error {
+func (t *Tombstoner) AddRange(keys [][]byte, min, max int64) error {
 	if len(keys) == 0 {
 		return nil
 	}
@@ -164,7 +166,10 @@ func (t *Tombstoner) Walk(fn func(t Tombstone) error) error {
 		return err
 	}
 
-	if binary.BigEndian.Uint32(b[:]) == v2header {
+	header := binary.BigEndian.Uint32(b[:])
+	if header == v3header {
+		return t.readTombstoneV3(f, fn)
+	} else if header == v2header {
 		return t.readTombstoneV2(f, fn)
 	}
 	return t.readTombstoneV1(f, fn)
@@ -181,30 +186,36 @@ func (t *Tombstoner) writeTombstone(tombstones []Tombstone) error {
 
 	bw := bufio.NewWriterSize(tmp, 1024*1024)
 
-	binary.BigEndian.PutUint32(b[:4], v2header)
+	binary.BigEndian.PutUint32(b[:4], v3header)
 	if _, err := bw.Write(b[:4]); err != nil {
 		return err
 	}
 
+	gz := gzip.NewWriter(bw)
+
 	for _, t := range tombstones {
 		binary.BigEndian.PutUint32(b[:4], uint32(len(t.Key)))
-		if _, err := bw.Write(b[:4]); err != nil {
+		if _, err := gz.Write(b[:4]); err != nil {
 			return err
 		}
-		if _, err := bw.WriteString(t.Key); err != nil {
+		if _, err := gz.Write([]byte(t.Key)); err != nil {
 			return err
 		}
 		binary.BigEndian.PutUint64(b[:], uint64(t.Min))
-		if _, err := bw.Write(b[:]); err != nil {
+		if _, err := gz.Write(b[:]); err != nil {
 			return err
 		}
 
 		binary.BigEndian.PutUint64(b[:], uint64(t.Max))
-		if _, err := bw.Write(b[:]); err != nil {
+		if _, err := gz.Write(b[:]); err != nil {
 			return err
 		}
 	}
 
+	if err := gz.Close(); err != nil {
+		return err
+	}
+
 	if err := bw.Flush(); err != nil {
 		return err
 	}
@@ -248,7 +259,7 @@ func (t *Tombstoner) readTombstoneV1(f *os.File, fn func(t Tombstone) error) err
 			continue
 		}
 		if err := fn(Tombstone{
-			Key: line,
+			Key: []byte(line),
 			Min: math.MinInt64,
 			Max: math.MaxInt64,
 		}); err != nil {
@@ -263,7 +274,7 @@ func (t *Tombstoner) readTombstoneV1(f *os.File, fn func(t Tombstone) error) err
 // format is binary.
 func (t *Tombstoner) readTombstoneV2(f *os.File, fn func(t Tombstone) error) error {
 	// Skip header, already checked earlier
-	if _, err := f.Seek(v2headerSize, io.SeekStart); err != nil {
+	if _, err := f.Seek(headerSize, io.SeekStart); err != nil {
 		return err
 	}
 	n := int64(4)
@@ -276,7 +287,7 @@ func (t *Tombstoner) readTombstoneV2(f *os.File, fn func(t Tombstone) error) err
 
 	var (
 		min, max int64
-		key      string
+		key      []byte
 	)
 	b := make([]byte, 4096)
 	for {
@@ -297,7 +308,7 @@ func (t *Tombstoner) readTombstoneV2(f *os.File, fn func(t Tombstone) error) err
 		if _, err := f.Read(b[:keyLen]); err != nil {
 			return err
 		}
-		key = string(b[:keyLen])
+		key = b[:keyLen]
 		n += int64(keyLen)
 
 		if _, err := f.Read(b[:8]); err != nil {
@@ -323,6 +334,69 @@ func (t *Tombstoner) readTombstoneV2(f *os.File, fn func(t Tombstone) error) err
 	}
 }
 
+// readTombstoneV3 reads the third version of tombstone files that are capable
+// of storing keys and the range of time for the key that points were deleted. This
+// format is a binary and compressed with gzip.
+func (t *Tombstoner) readTombstoneV3(f *os.File, fn func(t Tombstone) error) error {
+	// Skip header, already checked earlier
+	if _, err := f.Seek(headerSize, io.SeekStart); err != nil {
+		return err
+	}
+
+	var (
+		min, max int64
+		key      []byte
+	)
+
+	gr, err := gzip.NewReader(bufio.NewReader(f))
+	if err != nil {
+		return err
+	}
+	defer gr.Close()
+
+	b := make([]byte, 4096)
+	for {
+		if _, err = io.ReadFull(gr, b[:4]); err == io.EOF || err == io.ErrUnexpectedEOF {
+			return nil
+		} else if err != nil {
+			return err
+		}
+
+		keyLen := int(binary.BigEndian.Uint32(b[:4]))
+		if keyLen > len(b) {
+			b = make([]byte, keyLen)
+		}
+
+		if _, err := io.ReadFull(gr, b[:keyLen]); err != nil {
+			return err
+		}
+
+		// Copy the key since b is re-used
+		key = make([]byte, keyLen)
+		copy(key, b[:keyLen])
+
+		if _, err := io.ReadFull(gr, b[:8]); err != nil {
+			return err
+		}
+
+		min = int64(binary.BigEndian.Uint64(b[:8]))
+
+		if _, err := io.ReadFull(gr, b[:8]); err != nil {
+			return err
+		}
+
+		max = int64(binary.BigEndian.Uint64(b[:8]))
+
+		if err := fn(Tombstone{
+			Key: key,
+			Min: min,
+			Max: max,
+		}); err != nil {
+			return err
+		}
+	}
+}
+
 func (t *Tombstoner) tombstonePath() string {
 	if strings.HasSuffix(t.Path, "tombstone") {
 		return t.Path
