@@ -71,7 +71,7 @@ import (
 	"io"
 	"os"
 	"sort"
-	"sync"
+	"strings"
 	"time"
 )
 
@@ -80,6 +80,7 @@ const (
 	// identify the file as a tsm1 formatted file
 	MagicNumber uint32 = 0x16D116D1
 
+	// Version indicates the version of the TSM file format.
 	Version byte = 1
 
 	// Size in bytes of an index entry
@@ -96,70 +97,86 @@ const (
 
 	// max length of a key in an index entry (measurement + tags)
 	maxKeyLength = (1 << (2 * 8)) - 1
+
+	// The threshold amount data written before we periodically fsync a TSM file.  This helps avoid
+	// long pauses due to very large fsyncs at the end of writing a TSM file.
+	fsyncEvery = 25 * 1024 * 1024
 )
 
 var (
-	ErrNoValues             = fmt.Errorf("no values written")
-	ErrTSMClosed            = fmt.Errorf("tsm file closed")
+	//ErrNoValues is returned when TSMWriter.WriteIndex is called and there are no values to write.
+	ErrNoValues = fmt.Errorf("no values written")
+
+	// ErrTSMClosed is returned when performing an operation against a closed TSM file.
+	ErrTSMClosed = fmt.Errorf("tsm file closed")
+
+	// ErrMaxKeyLengthExceeded is returned when attempting to write a key that is too long.
 	ErrMaxKeyLengthExceeded = fmt.Errorf("max key length exceeded")
-	ErrMaxBlocksExceeded    = fmt.Errorf("max blocks exceeded")
+
+	// ErrMaxBlocksExceeded is returned when attempting to write a block past the allowed number.
+	ErrMaxBlocksExceeded = fmt.Errorf("max blocks exceeded")
 )
 
 // TSMWriter writes TSM formatted key and values.
 type TSMWriter interface {
 	// Write writes a new block for key containing and values.  Writes append
 	// blocks in the order that the Write function is called.  The caller is
-	// responsible for ensuring keys and blocks or sorted appropriately.
+	// responsible for ensuring keys and blocks are sorted appropriately.
 	// Values are encoded as a full block.  The caller is responsible for
-	// ensuring a fixed number of values are encoded in each block as wells as
+	// ensuring a fixed number of values are encoded in each block as well as
 	// ensuring the Values are sorted. The first and last timestamp values are
 	// used as the minimum and maximum values for the index entry.
-	Write(key string, values Values) error
+	Write(key []byte, values Values) error
 
 	// WriteBlock writes a new block for key containing the bytes in block.  WriteBlock appends
 	// blocks in the order that the WriteBlock function is called.  The caller is
 	// responsible for ensuring keys and blocks are sorted appropriately, and that the
 	// block and index information is correct for the block.  The minTime and maxTime
 	// timestamp values are used as the minimum and maximum values for the index entry.
-	WriteBlock(key string, minTime, maxTime int64, block []byte) error
+	WriteBlock(key []byte, minTime, maxTime int64, block []byte) error
 
 	// WriteIndex finishes the TSM write streams and writes the index.
 	WriteIndex() error
 
-	// Closes any underlying file resources.
+	// Flushes flushes all pending changes to the underlying file resources.
+	Flush() error
+
+	// Close closes any underlying file resources.
 	Close() error
 
-	// Size returns the current size in bytes of the file
+	// Size returns the current size in bytes of the file.
 	Size() uint32
+
+	Remove() error
 }
 
-// IndexWriter writes a TSMIndex
+// IndexWriter writes a TSMIndex.
 type IndexWriter interface {
 	// Add records a new block entry for a key in the index.
-	Add(key string, blockType byte, minTime, maxTime int64, offset int64, size uint32)
+	Add(key []byte, blockType byte, minTime, maxTime int64, offset int64, size uint32)
 
 	// Entries returns all index entries for a key.
-	Entries(key string) []IndexEntry
-
-	// Keys returns the unique set of keys in the index.
-	Keys() []string
+	Entries(key []byte) []IndexEntry
 
 	// KeyCount returns the count of unique keys in the index.
 	KeyCount() int
 
-	// Size returns the size of a the current index in bytes
+	// Size returns the size of a the current index in bytes.
 	Size() uint32
 
 	// MarshalBinary returns a byte slice encoded version of the index.
 	MarshalBinary() ([]byte, error)
 
-	// WriteTo writes the index contents to a writer
+	// WriteTo writes the index contents to a writer.
 	WriteTo(w io.Writer) (int64, error)
+
+	Close() error
+
+	Remove() error
 }
 
 // IndexEntry is the index information for a given block in a TSM file.
 type IndexEntry struct {
-
 	// The min and max time of all points stored in the block.
 	MinTime, MaxTime int64
 
@@ -170,10 +187,10 @@ type IndexEntry struct {
 	Size uint32
 }
 
-// UnmarshalBinary decodes an IndexEntry from a byte slice
+// UnmarshalBinary decodes an IndexEntry from a byte slice.
 func (e *IndexEntry) UnmarshalBinary(b []byte) error {
-	if len(b) != indexEntrySize {
-		return fmt.Errorf("unmarshalBinary: short buf: %v != %v", indexEntrySize, len(b))
+	if len(b) < indexEntrySize {
+		return fmt.Errorf("unmarshalBinary: short buf: %v < %v", len(b), indexEntrySize)
 	}
 	e.MinTime = int64(binary.BigEndian.Uint64(b[:8]))
 	e.MaxTime = int64(binary.BigEndian.Uint64(b[8:16]))
@@ -182,8 +199,8 @@ func (e *IndexEntry) UnmarshalBinary(b []byte) error {
 	return nil
 }
 
-// AppendTo will write a binary-encoded version of IndexEntry to b, allocating
-// and returning a new slice, if necessary
+// AppendTo writes a binary-encoded version of IndexEntry to b, allocating
+// and returning a new slice, if necessary.
 func (e *IndexEntry) AppendTo(b []byte) []byte {
 	if len(b) < indexEntrySize {
 		if cap(b) < indexEntrySize {
@@ -201,81 +218,149 @@ func (e *IndexEntry) AppendTo(b []byte) []byte {
 	return b
 }
 
-// Returns true if this IndexEntry may contain values for the given time.  The min and max
-// times are inclusive.
+// Contains returns true if this IndexEntry may contain values for the given time.
+// The min and max times are inclusive.
 func (e *IndexEntry) Contains(t int64) bool {
 	return e.MinTime <= t && e.MaxTime >= t
 }
 
+// OverlapsTimeRange returns true if the given time ranges are completely within the entry's time bounds.
 func (e *IndexEntry) OverlapsTimeRange(min, max int64) bool {
 	return e.MinTime <= max && e.MaxTime >= min
 }
 
+// String returns a string representation of the entry.
 func (e *IndexEntry) String() string {
 	return fmt.Sprintf("min=%s max=%s ofs=%d siz=%d",
 		time.Unix(0, e.MinTime).UTC(), time.Unix(0, e.MaxTime).UTC(), e.Offset, e.Size)
 }
 
+// NewIndexWriter returns a new IndexWriter.
 func NewIndexWriter() IndexWriter {
-	return &directIndex{
-		blocks: map[string]*indexEntries{},
-	}
+	buf := bytes.NewBuffer(make([]byte, 0, 1024*1024))
+	return &directIndex{buf: buf, w: bufio.NewWriter(buf)}
+}
+
+// NewIndexWriter returns a new IndexWriter.
+func NewDiskIndexWriter(f *os.File) IndexWriter {
+	return &directIndex{fd: f, w: bufio.NewWriterSize(f, 1024*1024)}
+}
+
+// indexBlock represent an index information for a series within a TSM file.
+type indexBlock struct {
+	key     []byte
+	entries *indexEntries
+}
+
+type syncer interface {
+	Name() string
+	Sync() error
 }
 
 // directIndex is a simple in-memory index implementation for a TSM file.  The full index
 // must fit in memory.
 type directIndex struct {
-	mu     sync.RWMutex
-	size   uint32
-	blocks map[string]*indexEntries
+	keyCount int
+	size     uint32
+
+	// The bytes written count of when we last fsync'd
+	lastSync uint32
+	fd       *os.File
+	buf      *bytes.Buffer
+
+	f syncer
+
+	w *bufio.Writer
+
+	key          []byte
+	indexEntries *indexEntries
 }
 
-func (d *directIndex) Add(key string, blockType byte, minTime, maxTime int64, offset int64, size uint32) {
-	d.mu.Lock()
-	defer d.mu.Unlock()
-
-	entries := d.blocks[key]
-	if entries == nil {
-		entries = &indexEntries{
-			Type: blockType,
-		}
-		d.blocks[key] = entries
+func (d *directIndex) Add(key []byte, blockType byte, minTime, maxTime int64, offset int64, size uint32) {
+	// Is this the first block being added?
+	if len(d.key) == 0 {
 		// size of the key stored in the index
 		d.size += uint32(2 + len(key))
-
 		// size of the count of entries stored in the index
 		d.size += indexCountSize
-	}
-	entries.entries = append(entries.entries, IndexEntry{
-		MinTime: minTime,
-		MaxTime: maxTime,
-		Offset:  offset,
-		Size:    size,
-	})
 
-	// size of the encoded index entry
-	d.size += indexEntrySize
+		d.key = key
+		if d.indexEntries == nil {
+			d.indexEntries = &indexEntries{}
+		}
+		d.indexEntries.Type = blockType
+		d.indexEntries.entries = append(d.indexEntries.entries, IndexEntry{
+			MinTime: minTime,
+			MaxTime: maxTime,
+			Offset:  offset,
+			Size:    size,
+		})
+
+		// size of the encoded index entry
+		d.size += indexEntrySize
+		d.keyCount++
+		return
+	}
+
+	// See if were still adding to the same series key.
+	cmp := bytes.Compare(d.key, key)
+	if cmp == 0 {
+		// The last block is still this key
+		d.indexEntries.entries = append(d.indexEntries.entries, IndexEntry{
+			MinTime: minTime,
+			MaxTime: maxTime,
+			Offset:  offset,
+			Size:    size,
+		})
+
+		// size of the encoded index entry
+		d.size += indexEntrySize
+
+	} else if cmp < 0 {
+		d.flush(d.w)
+		// We have a new key that is greater than the last one so we need to add
+		// a new index block section.
+
+		// size of the key stored in the index
+		d.size += uint32(2 + len(key))
+		// size of the count of entries stored in the index
+		d.size += indexCountSize
+
+		d.key = key
+		d.indexEntries.Type = blockType
+		d.indexEntries.entries = append(d.indexEntries.entries, IndexEntry{
+			MinTime: minTime,
+			MaxTime: maxTime,
+			Offset:  offset,
+			Size:    size,
+		})
+
+		// size of the encoded index entry
+		d.size += indexEntrySize
+		d.keyCount++
+	} else {
+		// Keys can't be added out of order.
+		panic(fmt.Sprintf("keys must be added in sorted order: %s < %s", string(key), string(d.key)))
+	}
 }
 
-func (d *directIndex) entries(key string) []IndexEntry {
-	entries := d.blocks[key]
-	if entries == nil {
+func (d *directIndex) entries(key []byte) []IndexEntry {
+	if len(d.key) == 0 {
 		return nil
 	}
-	return entries.entries
+
+	if bytes.Equal(d.key, key) {
+		return d.indexEntries.entries
+	}
+
+	return nil
 }
 
-func (d *directIndex) Entries(key string) []IndexEntry {
-	d.mu.RLock()
-	defer d.mu.RUnlock()
-
+func (d *directIndex) Entries(key []byte) []IndexEntry {
 	return d.entries(key)
 }
 
-func (d *directIndex) Entry(key string, t int64) *IndexEntry {
-	d.mu.RLock()
-	defer d.mu.RUnlock()
-
+func (d *directIndex) Entry(key []byte, t int64) *IndexEntry {
 	entries := d.entries(key)
 	for _, entry := range entries {
 		if entry.Contains(t) {
@@ -285,45 +370,73 @@ func (d *directIndex) Entry(key string, t int64) *IndexEntry {
 	return nil
 }
 
-func (d *directIndex) Keys() []string {
-	d.mu.RLock()
-	defer d.mu.RUnlock()
-
-	var keys []string
-	for k := range d.blocks {
-		keys = append(keys, k)
-	}
-	sort.Strings(keys)
-	return keys
-}
-
 func (d *directIndex) KeyCount() int {
-	d.mu.RLock()
-	n := len(d.blocks)
-	d.mu.RUnlock()
-	return n
+	return d.keyCount
 }
 
-func (d *directIndex) addEntries(key string, entries *indexEntries) {
-	existing := d.blocks[key]
-	if existing == nil {
-		d.blocks[key] = entries
-		return
+// copyBuffer is the actual implementation of Copy and CopyBuffer.
+// if buf is nil, one is allocated.  This is copied from the Go stdlib
+// in order to remove the fast path WriteTo calls which circumvent any
+// IO throttling as well as to add periodic fsyncs to avoid long stalls.
+func copyBuffer(f syncer, dst io.Writer, src io.Reader, buf []byte) (written int64, err error) {
+	if buf == nil {
+		buf = make([]byte, 32*1024)
 	}
-	existing.entries = append(existing.entries, entries.entries...)
+	var lastSync int64
+	for {
+		nr, er := src.Read(buf)
+		if nr > 0 {
+			nw, ew := dst.Write(buf[0:nr])
+			if nw > 0 {
+				written += int64(nw)
+			}
+
+			if written-lastSync > fsyncEvery {
+				if err := f.Sync(); err != nil {
+					return 0, err
+				}
+				lastSync = written
+			}
+			if ew != nil {
+				err = ew
+				break
+			}
+			if nr != nw {
+				err = io.ErrShortWrite
+				break
+			}
+		}
+		if er != nil {
+			if er != io.EOF {
+				err = er
+			}
+			break
+		}
+	}
+	return written, err
 }
 
 func (d *directIndex) WriteTo(w io.Writer) (int64, error) {
-	d.mu.RLock()
-	defer d.mu.RUnlock()
-
-	// Index blocks are writtens sorted by key
-	keys := make([]string, 0, len(d.blocks))
-	for k := range d.blocks {
-		keys = append(keys, k)
+	if _, err := d.flush(d.w); err != nil {
+		return 0, err
 	}
-	sort.Strings(keys)
 
+	if err := d.w.Flush(); err != nil {
+		return 0, err
+	}
+
+	if d.fd == nil {
+		return copyBuffer(d.f, w, d.buf, nil)
+	}
+
+	if _, err := d.fd.Seek(0, io.SeekStart); err != nil {
+		return 0, err
+	}
+
+	return io.Copy(w, bufio.NewReaderSize(d.fd, 1024*1024))
+}
+
+func (d *directIndex) flush(w io.Writer) (int64, error) {
 	var (
 		n   int
 		err error
@@ -331,45 +444,64 @@ func (d *directIndex) WriteTo(w io.Writer) (int64, error) {
 		N   int64
 	)
 
-	// For each key, individual entries are sorted by time
-	for _, key := range keys {
-		entries := d.blocks[key]
-
-		if entries.Len() > maxIndexEntries {
-			return N, fmt.Errorf("key '%s' exceeds max index entries: %d > %d", key, entries.Len(), maxIndexEntries)
-		}
-		sort.Sort(entries)
-
-		binary.BigEndian.PutUint16(buf[0:2], uint16(len(key)))
-		buf[2] = entries.Type
-		binary.BigEndian.PutUint16(buf[3:5], uint16(entries.Len()))
-
-		// Append the key length and key
-		if n, err = w.Write(buf[0:2]); err != nil {
-			return int64(n) + N, fmt.Errorf("write: writer key length error: %v", err)
-		}
-		N += int64(n)
-
-		if n, err = io.WriteString(w, key); err != nil {
-			return int64(n) + N, fmt.Errorf("write: writer key error: %v", err)
-		}
-		N += int64(n)
-
-		// Append the block type and count
-		if n, err = w.Write(buf[2:5]); err != nil {
-			return int64(n) + N, fmt.Errorf("write: writer block type and count error: %v", err)
-		}
-		N += int64(n)
-
-		// Append each index entry for all blocks for this key
-		var n64 int64
-		if n64, err = entries.WriteTo(w); err != nil {
-			return n64 + N, fmt.Errorf("write: writer entries error: %v", err)
-		}
-		N += n64
-
+	if len(d.key) == 0 {
+		return 0, nil
 	}
+	// For each key, individual entries are sorted by time
+	key := d.key
+	entries := d.indexEntries
+
+	if entries.Len() > maxIndexEntries {
+		return N, fmt.Errorf("key '%s' exceeds max index entries: %d > %d", key, entries.Len(), maxIndexEntries)
+	}
+
+	if !sort.IsSorted(entries) {
+		sort.Sort(entries)
+	}
+
+	binary.BigEndian.PutUint16(buf[0:2], uint16(len(key)))
+	buf[2] = entries.Type
+	binary.BigEndian.PutUint16(buf[3:5], uint16(entries.Len()))
+
+	// Append the key length and key
+	if n, err = w.Write(buf[0:2]); err != nil {
+		return int64(n) + N, fmt.Errorf("write: writer key length error: %v", err)
+	}
+	N += int64(n)
+
+	if n, err = w.Write(key); err != nil {
+		return int64(n) + N, fmt.Errorf("write: writer key error: %v", err)
+	}
+	N += int64(n)
+
+	// Append the block type and count
+	if n, err = w.Write(buf[2:5]); err != nil {
+		return int64(n) + N, fmt.Errorf("write: writer block type and count error: %v", err)
+	}
+	N += int64(n)
+
+	// Append each index entry for all blocks for this key
+	var n64 int64
+	if n64, err = entries.WriteTo(w); err != nil {
+		return n64 + N, fmt.Errorf("write: writer entries error: %v", err)
+	}
+	N += n64
+
+	d.key = nil
+	d.indexEntries.Type = 0
+	d.indexEntries.entries = d.indexEntries.entries[:0]
+
+	// If this is a disk based index and we've written more than the fsync threshold,
+	// fsync the data to avoid long pauses later on.
+	if d.fd != nil && d.size-d.lastSync > fsyncEvery {
+		if err := d.fd.Sync(); err != nil {
+			return N, err
+		}
+		d.lastSync = d.size
+	}
+
 	return N, nil
+
 }
 
 func (d *directIndex) MarshalBinary() ([]byte, error) {
@@ -380,50 +512,73 @@ func (d *directIndex) MarshalBinary() ([]byte, error) {
 	return b.Bytes(), nil
 }
 
-func (d *directIndex) UnmarshalBinary(b []byte) error {
-	d.mu.Lock()
-	defer d.mu.Unlock()
-
-	d.size = uint32(len(b))
-
-	var pos int
-	for pos < len(b) {
-		n, key, err := readKey(b[pos:])
-		if err != nil {
-			return fmt.Errorf("readIndex: read key error: %v", err)
-		}
-		pos += n
-
-		var entries indexEntries
-		n, err = readEntries(b[pos:], &entries)
-		if err != nil {
-			return fmt.Errorf("readIndex: read entries error: %v", err)
-		}
-
-		pos += n
-		d.addEntries(string(key), &entries)
-	}
-	return nil
-}
-
 func (d *directIndex) Size() uint32 {
 	return d.size
 }
 
+func (d *directIndex) Close() error {
+	// Flush anything remaining in the index
+	if err := d.w.Flush(); err != nil {
+		return err
+	}
+
+	if d.fd == nil {
+		return nil
+	}
+
+	if err := d.fd.Close(); err != nil {
+		return err
+	}
+	return os.Remove(d.fd.Name())
+}
+
+// Remove removes the index from any tempory storage
+func (d *directIndex) Remove() error {
+	if d.fd == nil {
+		return nil
+	}
+
+	// Close the file handle to prevent leaking.  We ignore the error because
+	// we just want to cleanup and remove the file.
+	_ = d.fd.Close()
+
+	return os.Remove(d.fd.Name())
+}
+
 // tsmWriter writes keys and values in the TSM format
 type tsmWriter struct {
 	wrapped io.Writer
 	w       *bufio.Writer
 	index   IndexWriter
 	n       int64
+
+	// The bytes written count of when we last fsync'd
+	lastSync int64
 }
 
+// NewTSMWriter returns a new TSMWriter writing to w.
 func NewTSMWriter(w io.Writer) (TSMWriter, error) {
-	index := &directIndex{
-		blocks: map[string]*indexEntries{},
+	index := NewIndexWriter()
+	return &tsmWriter{wrapped: w, w: bufio.NewWriterSize(w, 1024*1024), index: index}, nil
+}
+
+// NewTSMWriterWithDiskBuffer returns a new TSMWriter writing to w and will use a disk
+// based buffer for the TSM index if possible.
+func NewTSMWriterWithDiskBuffer(w io.Writer) (TSMWriter, error) {
+	var index IndexWriter
+	// Make sure is a File so we can write the temp index alongside it.
+	if fw, ok := w.(syncer); ok {
+		f, err := os.OpenFile(strings.TrimSuffix(fw.Name(), ".tsm.tmp")+".idx.tmp", os.O_CREATE|os.O_RDWR|os.O_EXCL, 0666)
+		if err != nil {
+			return nil, err
+		}
+		index = NewDiskIndexWriter(f)
+	} else {
+		// w is not a file, just use an inmem index
+		index = NewIndexWriter()
 	}
 
-	return &tsmWriter{wrapped: w, w: bufio.NewWriterSize(w, 4*1024*1024), index: index}, nil
+	return &tsmWriter{wrapped: w, w: bufio.NewWriterSize(w, 1024*1024), index: index}, nil
 }
 
 func (t *tsmWriter) writeHeader() error {
@@ -439,16 +594,17 @@ func (t *tsmWriter) writeHeader() error {
 	return nil
 }
 
-func (t *tsmWriter) Write(key string, values Values) error {
+// Write writes a new block containing key and values.
+func (t *tsmWriter) Write(key []byte, values Values) error {
+	if len(key) > maxKeyLength {
+		return ErrMaxKeyLengthExceeded
+	}
+
 	// Nothing to write
 	if len(values) == 0 {
 		return nil
 	}
 
-	if len(key) > maxKeyLength {
-		return ErrMaxKeyLengthExceeded
-	}
-
 	// Write header only after we have some data to write.
 	if t.n == 0 {
 		if err := t.writeHeader(); err != nil {
@@ -491,7 +647,11 @@ func (t *tsmWriter) Write(key string, values Values) error {
 // WriteBlock writes block for the given key and time range to the TSM file.  If the write
 // exceeds max entries for a given key, ErrMaxBlocksExceeded is returned.  This indicates
 // that the index is now full for this key and no future writes to this key will succeed.
-func (t *tsmWriter) WriteBlock(key string, minTime, maxTime int64, block []byte) error {
+func (t *tsmWriter) WriteBlock(key []byte, minTime, maxTime int64, block []byte) error {
+	if len(key) > maxKeyLength {
+		return ErrMaxKeyLengthExceeded
+	}
+
 	// Nothing to write
 	if len(block) == 0 {
 		return nil
@@ -529,6 +689,14 @@ func (t *tsmWriter) WriteBlock(key string, minTime, maxTime int64, block []byte)
 	// Increment file position pointer (checksum + block len)
 	t.n += int64(n)
 
+	// fsync the file periodically to avoid long pauses with very big files.
+	if t.n-t.lastSync > fsyncEvery {
+		if err := t.sync(); err != nil {
+			return err
+		}
+		t.lastSync = t.n
+	}
+
 	if len(t.index.Entries(key)) >= maxIndexEntries {
 		return ErrMaxBlocksExceeded
 	}
@@ -537,7 +705,7 @@ func (t *tsmWriter) WriteBlock(key string, minTime, maxTime int64, block []byte)
 }
 
 // WriteIndex writes the index section of the file.  If there are no index entries to write,
-// this returns ErrNoValues
+// this returns ErrNoValues.
 func (t *tsmWriter) WriteIndex() error {
 	indexPos := t.n
 
@@ -545,6 +713,12 @@ func (t *tsmWriter) WriteIndex() error {
 		return ErrNoValues
 	}
 
+	// Set the destination file on the index so we can periodically
+	// fsync while writing the index.
+	if f, ok := t.wrapped.(syncer); ok {
+		t.index.(*directIndex).f = f
+	}
+
 	// Write the index
 	if _, err := t.index.WriteTo(t.w); err != nil {
 		return err
@@ -558,16 +732,31 @@ func (t *tsmWriter) WriteIndex() error {
 	return err
 }
 
-func (t *tsmWriter) Close() error {
+func (t *tsmWriter) Flush() error {
 	if err := t.w.Flush(); err != nil {
 		return err
 	}
 
+	return t.sync()
+}
+
+func (t *tsmWriter) sync() error {
 	if f, ok := t.wrapped.(*os.File); ok {
 		if err := f.Sync(); err != nil {
 			return err
 		}
 	}
+	return nil
+}
+
+func (t *tsmWriter) Close() error {
+	if err := t.Flush(); err != nil {
+		return err
+	}
+
+	if err := t.index.Close(); err != nil {
+		return err
+	}
 
 	if c, ok := t.wrapped.(io.Closer); ok {
 		return c.Close()
@@ -575,11 +764,27 @@ func (t *tsmWriter) Close() error {
 	return nil
 }
 
+// Remove removes any temporary storage used by the writer.
+func (t *tsmWriter) Remove() error {
+	if err := t.index.Remove(); err != nil {
+		return err
+	}
+
+	if f, ok := t.wrapped.(*os.File); ok {
+		// Close the file handle to prevent leaking.  We ignore the error because
+		// we just want to cleanup and remove the file.
+		_ = f.Close()
+
+		return os.Remove(f.Name())
+	}
+	return nil
+}
+
 func (t *tsmWriter) Size() uint32 {
 	return uint32(t.n) + t.index.Size()
 }
 
-// verifyVersion will verify that the reader's bytes are a TSM byte
+// verifyVersion verifies that the reader's bytes are a TSM byte
 // stream of the correct version (1)
 func verifyVersion(r io.ReadSeeker) error {
 	_, err := r.Seek(0, 0)
