@@ -6,25 +6,32 @@ import (
 	"fmt"
 	"regexp"
 
-	"github.com/influxdata/influxdb/influxql"
 	"github.com/influxdata/influxdb/models"
 	"github.com/influxdata/influxdb/pkg/bloom"
 	"github.com/influxdata/influxdb/pkg/bytesutil"
 	"github.com/influxdata/influxdb/pkg/estimator"
 	"github.com/influxdata/influxdb/pkg/estimator/hll"
+	"github.com/influxdata/influxdb/query"
 	"github.com/influxdata/influxdb/tsdb"
+	"github.com/influxdata/influxql"
 )
 
 // FileSet represents a collection of files.
 type FileSet struct {
-	levels  []CompactionLevel
-	files   []File
-	filters []*bloom.Filter // per-level filters
+	levels   []CompactionLevel
+	files    []File
+	filters  []*bloom.Filter // per-level filters
+	database string
 }
 
 // NewFileSet returns a new instance of FileSet.
-func NewFileSet(levels []CompactionLevel, files []File) (*FileSet, error) {
-	fs := &FileSet{levels: levels, files: files}
+func NewFileSet(database string, levels []CompactionLevel, files []File) (*FileSet, error) {
+	fs := &FileSet{
+		levels:   levels,
+		files:    files,
+		filters:  make([]*bloom.Filter, len(levels)),
+		database: database,
+	}
 	if err := fs.buildFilters(); err != nil {
 		return nil, err
 	}
@@ -32,9 +39,9 @@ func NewFileSet(levels []CompactionLevel, files []File) (*FileSet, error) {
 }
 
 // Close closes all the files in the file set.
-func (p FileSet) Close() error {
+func (fs FileSet) Close() error {
 	var err error
-	for _, f := range p.files {
+	for _, f := range fs.files {
 		if e := f.Close(); e != nil && err == nil {
 			err = e
 		}
@@ -56,13 +63,19 @@ func (fs *FileSet) Release() {
 	}
 }
 
-// Prepend returns a new file set with f added at the beginning.
-func (fs *FileSet) Prepend(f File) (*FileSet, error) {
-	return NewFileSet(fs.levels, append([]File{f}, fs.files...))
+// PrependLogFile returns a new file set with f added at the beginning.
+// Filters do not need to be rebuilt because log files have no bloom filter.
+func (fs *FileSet) PrependLogFile(f *LogFile) *FileSet {
+	return &FileSet{
+		database: fs.database,
+		levels:   fs.levels,
+		files:    append([]File{f}, fs.files...),
+		filters:  fs.filters,
+	}
 }
 
 // MustReplace swaps a list of files for a single file and returns a new file set.
-// The caller should always guarentee that the files exist and are contiguous.
+// The caller should always guarantee that the files exist and are contiguous.
 func (fs *FileSet) MustReplace(oldFiles []File, newFile File) *FileSet {
 	assert(len(oldFiles) > 0, "cannot replace empty files")
 
@@ -89,11 +102,27 @@ func (fs *FileSet) MustReplace(oldFiles []File, newFile File) *FileSet {
 	other[i] = newFile
 	copy(other[i+1:], fs.files[i+len(oldFiles):])
 
-	fs, err := NewFileSet(fs.levels, other)
-	if err != nil {
+	// Copy existing bloom filters.
+	filters := make([]*bloom.Filter, len(fs.filters))
+	// copy(filters, fs.filters)
+
+	// Clear filters at replaced file levels.
+	filters[newFile.Level()] = nil
+	for _, f := range oldFiles {
+		filters[f.Level()] = nil
+	}
+
+	// Build new fileset and rebuild changed filters.
+	newFS := &FileSet{
+		levels:   fs.levels,
+		files:    other,
+		filters:  filters,
+		database: fs.database,
+	}
+	if err := newFS.buildFilters(); err != nil {
 		panic("cannot build file set: " + err.Error())
 	}
-	return fs
+	return newFS
 }
 
 // MaxID returns the highest file identifier.
@@ -223,6 +252,17 @@ func (fs *FileSet) TagKeyIterator(name []byte) TagKeyIterator {
 
 // MeasurementTagKeysByExpr extracts the tag keys wanted by the expression.
 func (fs *FileSet) MeasurementTagKeysByExpr(name []byte, expr influxql.Expr) (map[string]struct{}, error) {
+	// Return all keys if no condition was passed in.
+	if expr == nil {
+		m := make(map[string]struct{})
+		if itr := fs.TagKeyIterator(name); itr != nil {
+			for e := itr.Next(); e != nil; e = itr.Next() {
+				m[string(e.Key())] = struct{}{}
+			}
+		}
+		return m, nil
+	}
+
 	switch e := expr.(type) {
 	case *influxql.BinaryExpr:
 		switch e.Op {
@@ -288,7 +328,7 @@ func (fs *FileSet) MeasurementTagKeysByExpr(name []byte, expr influxql.Expr) (ma
 //
 // N.B tagValuesByKeyAndExpr relies on keys being sorted in ascending
 // lexicographic order.
-func (fs *FileSet) tagValuesByKeyAndExpr(name []byte, keys []string, expr influxql.Expr, fieldset *tsdb.MeasurementFieldSet) ([]map[string]struct{}, error) {
+func (fs *FileSet) tagValuesByKeyAndExpr(auth query.Authorizer, name []byte, keys []string, expr influxql.Expr, fieldset *tsdb.MeasurementFieldSet) ([]map[string]struct{}, error) {
 	itr, err := fs.seriesByExprIterator(name, expr, fieldset.Fields(string(name)))
 	if err != nil {
 		return nil, err
@@ -313,6 +353,9 @@ func (fs *FileSet) tagValuesByKeyAndExpr(name []byte, keys []string, expr influx
 
 	// Iterate all series to collect tag values.
 	for e := itr.Next(); e != nil; e = itr.Next() {
+		if auth != nil && !auth.AuthorizeSeriesRead(fs.database, e.Name(), e.Tags()) {
+			continue
+		}
 		for _, t := range e.Tags() {
 			if idx, ok := keyIdxs[string(t.Key)]; ok {
 				resultSet[idx][string(t.Value)] = struct{}{}
@@ -492,22 +535,28 @@ func (fs *FileSet) matchTagValueNotEqualNotEmptySeriesIterator(name, key []byte,
 	)
 }
 
-func (fs *FileSet) MeasurementNamesByExpr(expr influxql.Expr) ([][]byte, error) {
+func (fs *FileSet) MeasurementNamesByExpr(auth query.Authorizer, expr influxql.Expr) ([][]byte, error) {
 	// Return filtered list if expression exists.
 	if expr != nil {
-		return fs.measurementNamesByExpr(expr)
+		return fs.measurementNamesByExpr(auth, expr)
+	}
+
+	itr := fs.MeasurementIterator()
+	if itr == nil {
+		return nil, nil
 	}
 
 	// Iterate over all measurements if no condition exists.
 	var names [][]byte
-	itr := fs.MeasurementIterator()
 	for e := itr.Next(); e != nil; e = itr.Next() {
-		names = append(names, e.Name())
+		if fs.measurementAuthorizedSeries(auth, e.Name()) {
+			names = append(names, e.Name())
+		}
 	}
 	return names, nil
 }
 
-func (fs *FileSet) measurementNamesByExpr(expr influxql.Expr) ([][]byte, error) {
+func (fs *FileSet) measurementNamesByExpr(auth query.Authorizer, expr influxql.Expr) ([][]byte, error) {
 	if expr == nil {
 		return nil, nil
 	}
@@ -540,19 +589,19 @@ func (fs *FileSet) measurementNamesByExpr(expr influxql.Expr) ([][]byte, error)
 
 			// Match on name, if specified.
 			if tag.Val == "_name" {
-				return fs.measurementNamesByNameFilter(e.Op, value, regex), nil
+				return fs.measurementNamesByNameFilter(auth, e.Op, value, regex), nil
 			} else if influxql.IsSystemName(tag.Val) {
 				return nil, nil
 			}
-			return fs.measurementNamesByTagFilter(e.Op, tag.Val, value, regex), nil
+			return fs.measurementNamesByTagFilter(auth, e.Op, tag.Val, value, regex), nil
 
 		case influxql.OR, influxql.AND:
-			lhs, err := fs.measurementNamesByExpr(e.LHS)
+			lhs, err := fs.measurementNamesByExpr(auth, e.LHS)
 			if err != nil {
 				return nil, err
 			}
 
-			rhs, err := fs.measurementNamesByExpr(e.RHS)
+			rhs, err := fs.measurementNamesByExpr(auth, e.RHS)
 			if err != nil {
 				return nil, err
 			}
@@ -567,16 +616,20 @@ func (fs *FileSet) measurementNamesByExpr(expr influxql.Expr) ([][]byte, error)
 		}
 
 	case *influxql.ParenExpr:
-		return fs.measurementNamesByExpr(e.Expr)
+		return fs.measurementNamesByExpr(auth, e.Expr)
 	default:
 		return nil, fmt.Errorf("%#v", expr)
 	}
 }
 
 // measurementNamesByNameFilter returns matching measurement names in sorted order.
-func (fs *FileSet) measurementNamesByNameFilter(op influxql.Token, val string, regex *regexp.Regexp) [][]byte {
-	var names [][]byte
+func (fs *FileSet) measurementNamesByNameFilter(auth query.Authorizer, op influxql.Token, val string, regex *regexp.Regexp) [][]byte {
 	itr := fs.MeasurementIterator()
+	if itr == nil {
+		return nil
+	}
+
+	var names [][]byte
 	for e := itr.Next(); e != nil; e = itr.Next() {
 		var matched bool
 		switch op {
@@ -590,7 +643,7 @@ func (fs *FileSet) measurementNamesByNameFilter(op influxql.Token, val string, r
 			matched = !regex.Match(e.Name())
 		}
 
-		if matched {
+		if matched && fs.measurementAuthorizedSeries(auth, e.Name()) {
 			names = append(names, e.Name())
 		}
 	}
@@ -598,43 +651,82 @@ func (fs *FileSet) measurementNamesByNameFilter(op influxql.Token, val string, r
 	return names
 }
 
-func (fs *FileSet) measurementNamesByTagFilter(op influxql.Token, key, val string, regex *regexp.Regexp) [][]byte {
+func (fs *FileSet) measurementNamesByTagFilter(auth query.Authorizer, op influxql.Token, key, val string, regex *regexp.Regexp) [][]byte {
 	var names [][]byte
 
 	mitr := fs.MeasurementIterator()
+	if mitr == nil {
+		return nil
+	}
+
+	// valEqual determines if the provided []byte] is equal to the tag value
+	// to be filtered on.
+	valEqual := regex.Match
+	if op == influxql.EQ || op == influxql.NEQ {
+		vb := []byte(val)
+		valEqual = func(b []byte) bool { return bytes.Equal(vb, b) }
+	}
+
+	var tagMatch bool
+	var authorized bool
 	for me := mitr.Next(); me != nil; me = mitr.Next() {
-		// If the operator is non-regex, only check the specified value.
-		var tagMatch bool
-		if op == influxql.EQ || op == influxql.NEQ {
-			if fs.HasTagValue(me.Name(), []byte(key), []byte(val)) {
+		// If the measurement doesn't have the tag key, then it won't be considered.
+		if !fs.HasTagKey(me.Name(), []byte(key)) {
+			continue
+		}
+
+		tagMatch = false
+		// Authorization must be explicitly granted when an authorizer is present.
+		authorized = auth == nil
+
+		vitr := fs.TagValueIterator(me.Name(), []byte(key))
+		if vitr != nil {
+			for ve := vitr.Next(); ve != nil; ve = vitr.Next() {
+				if !valEqual(ve.Value()) {
+					continue
+				}
+
 				tagMatch = true
-			}
-		} else {
-			// Else, the operator is a regex and we have to check all tag
-			// values against the regular expression.
-			vitr := fs.TagValueIterator(me.Name(), []byte(key))
-			if vitr != nil {
-				for ve := vitr.Next(); ve != nil; ve = vitr.Next() {
-					if regex.Match(ve.Value()) {
-						tagMatch = true
+				if auth == nil {
+					break
+				}
+
+				// When an authorizer is present, the measurement should be
+				// included only if one of it's series is authorized.
+				sitr := fs.TagValueSeriesIterator(me.Name(), []byte(key), ve.Value())
+				if sitr == nil {
+					continue
+				}
+				// Locate a series with this matching tag value that's authorized.
+				for se := sitr.Next(); se != nil; se = sitr.Next() {
+					if auth.AuthorizeSeriesRead(fs.database, me.Name(), se.Tags()) {
+						authorized = true
 						break
 					}
 				}
+
+				if tagMatch && authorized {
+					// The measurement can definitely be included or rejected.
+					break
+				}
 			}
 		}
 
-		//
-		// XNOR gate
-		//
+		// For negation operators, to determine if the measurement is authorized,
+		// an authorized series belonging to the measurement must be located.
+		// Then, the measurement can be added iff !tagMatch && authorized.
+		if op == influxql.NEQ || op == influxql.NEQREGEX && !tagMatch {
+			authorized = fs.measurementAuthorizedSeries(auth, me.Name())
+		}
+
 		// tags match | operation is EQ | measurement matches
 		// --------------------------------------------------
 		//     True   |       True      |      True
 		//     True   |       False     |      False
 		//     False  |       True      |      False
 		//     False  |       False     |      True
-		if tagMatch == (op == influxql.EQ || op == influxql.EQREGEX) {
+		if tagMatch == (op == influxql.EQ || op == influxql.EQREGEX) && authorized {
 			names = append(names, me.Name())
-			continue
 		}
 	}
 
@@ -642,6 +734,22 @@ func (fs *FileSet) measurementNamesByTagFilter(op influxql.Token, key, val strin
 	return names
 }
 
+// measurementAuthorizedSeries determines if the measurement contains a series
+// that is authorized to be read.
+func (fs *FileSet) measurementAuthorizedSeries(auth query.Authorizer, name []byte) bool {
+	if auth == nil {
+		return true
+	}
+
+	sitr := fs.MeasurementSeriesIterator(name)
+	for series := sitr.Next(); series != nil; series = sitr.Next() {
+		if auth.AuthorizeSeriesRead(fs.database, name, series.Tags()) {
+			return true
+		}
+	}
+	return false
+}
+
 // HasSeries returns true if the series exists and is not tombstoned.
 func (fs *FileSet) HasSeries(name []byte, tags models.Tags, buf []byte) bool {
 	for _, f := range fs.files {
@@ -702,6 +810,7 @@ func (fs *FileSet) FilterNamesTags(names [][]byte, tagsSlice []models.Tags) ([][
 			newTagsSlice = append(newTagsSlice, tags)
 		}
 	}
+
 	return newNames, newTagsSlice
 }
 
@@ -736,7 +845,7 @@ func (fs *FileSet) MeasurementsSketches() (estimator.Sketch, estimator.Sketch, e
 // call is equivalent to MeasurementSeriesIterator().
 func (fs *FileSet) MeasurementSeriesByExprIterator(name []byte, expr influxql.Expr, fieldset *tsdb.MeasurementFieldSet) (SeriesIterator, error) {
 	// Return all series for the measurement if there are no tag expressions.
-	if expr == nil || influxql.OnlyTimeExpr(expr) {
+	if expr == nil {
 		return fs.MeasurementSeriesIterator(name), nil
 	}
 	return fs.seriesByExprIterator(name, expr, fieldset.CreateFieldsIfNotExists(name))
@@ -758,7 +867,9 @@ func (fs *FileSet) MeasurementSeriesKeysByExpr(name []byte, expr influxql.Expr,
 		// Check for unsupported field filters.
 		// Any remaining filters means there were fields (e.g., `WHERE value = 1.2`).
 		if e.Expr() != nil {
-			return nil, errors.New("fields not supported in WHERE clause during deletion")
+			if v, ok := e.Expr().(*influxql.BooleanLiteral); !ok || !v.Val {
+				return nil, errors.New("fields not supported in WHERE clause during deletion")
+			}
 		}
 
 		keys = append(keys, models.MakeKey(e.Name(), e.Tags()))
@@ -824,11 +935,6 @@ func (fs *FileSet) seriesByBinaryExprIterator(name []byte, n *influxql.BinaryExp
 		value = n.LHS
 	}
 
-	// For time literals, return all series and "true" as the filter.
-	if _, ok := value.(*influxql.TimeLiteral); ok || key.Val == "time" {
-		return newSeriesExprIterator(fs.MeasurementSeriesIterator(name), &influxql.BooleanLiteral{Val: true}), nil
-	}
-
 	// For fields, return all series from this measurement.
 	if key.Val != "_name" && ((key.Type == influxql.Unknown && mf.HasField(key.Val)) || key.Type == influxql.AnyField || (key.Type != influxql.Tag && key.Type != influxql.Unknown)) {
 		return newSeriesExprIterator(fs.MeasurementSeriesIterator(name), n), nil
@@ -918,31 +1024,38 @@ func (fs *FileSet) seriesByBinaryExprVarRefIterator(name, key []byte, value *inf
 // buildFilters builds a series existence filter for each compaction level.
 func (fs *FileSet) buildFilters() error {
 	if len(fs.levels) == 0 {
-		fs.filters = nil
 		return nil
 	}
 
-	// Generate filters for each level.
-	fs.filters = make([]*bloom.Filter, len(fs.levels))
+	// Move past log files (level=0).
+	files := fs.files
+	for len(files) > 0 && files[0].Level() == 0 {
+		files = files[1:]
+	}
 
-	// Merge filters at each level.
-	for _, f := range fs.files {
-		level := f.Level()
-
-		// Skip if file has no bloom filter.
-		if f.Filter() == nil {
+	// Build filters for each level where the filter is non-existent.
+	for level := range fs.levels {
+		// Clear filter if no files remain or next file is at a higher level.
+		if len(files) == 0 || files[0].Level() > level {
+			fs.filters[level] = nil
 			continue
 		}
 
-		// Initialize a filter if it doesn't exist.
-		if fs.filters[level] == nil {
-			lvl := fs.levels[level]
-			fs.filters[level] = bloom.NewFilter(lvl.M, lvl.K)
+		// Skip files at this level if filter already exists.
+		if fs.filters[level] != nil {
+			for len(files) > 0 && files[0].Level() == level {
+				files = files[1:]
+			}
+			continue
 		}
 
-		// Merge filter.
-		if err := fs.filters[level].Merge(f.Filter()); err != nil {
-			return err
+		// Build new filter from files at this level.
+		fs.filters[level] = bloom.NewFilter(fs.levels[level].M, fs.levels[level].K)
+		for len(files) > 0 && files[0].Level() == level {
+			if err := fs.filters[level].Merge(files[0].Filter()); err != nil {
+				return err
+			}
+			files = files[1:]
 		}
 	}
 
