@@ -479,7 +479,8 @@ type TagBlockEncoder struct {
 	trailer TagBlockTrailer
 
 	// Track tag keys.
-	keys []tagKeyEncodeEntry
+	keys      []tagKeyEncodeEntry
+	prevValue []byte
 }
 
 // NewTagBlockEncoder returns a new TagBlockEncoder.
@@ -527,6 +528,9 @@ func (enc *TagBlockEncoder) EncodeKey(key []byte, deleted bool) error {
 
 	enc.keys = append(enc.keys, entry)
 
+	// Clear previous value.
+	enc.prevValue = nil
+
 	return nil
 }
 
@@ -539,6 +543,13 @@ func (enc *TagBlockEncoder) EncodeValue(value []byte, deleted bool, seriesIDs []
 		return fmt.Errorf("zero length tag value not allowed")
 	}
 
+	// Validate that keys are in-order.
+	if cmp := bytes.Compare(enc.prevValue, value); cmp == 1 {
+		return fmt.Errorf("tag value out of order: prev=%s, new=%s", enc.prevValue, value)
+	} else if cmp == 0 {
+		return fmt.Errorf("tag value already encoded: %s", value)
+	}
+
 	// Save offset to hash map.
 	enc.offsets.Put(value, enc.n)
 
@@ -583,6 +594,9 @@ func (enc *TagBlockEncoder) EncodeValue(value []byte, deleted bool, seriesIDs []
 		return err
 	}
 
+	// Save previous value.
+	enc.prevValue = value
+
 	return nil
 }
 
