@@ -8,8 +8,9 @@ import (
 	"io"
 	"os"
 
-	"github.com/influxdata/influxdb/influxql"
 	"github.com/influxdata/influxdb/models"
+	"github.com/influxdata/influxdb/query"
+	"github.com/influxdata/influxql"
 )
 
 // LoadFactor is the fill percent for RHH indexes.
@@ -791,16 +792,6 @@ func (a byteSlices) Len() int           { return len(a) }
 func (a byteSlices) Swap(i, j int)      { a[i], a[j] = a[j], a[i] }
 func (a byteSlices) Less(i, j int) bool { return bytes.Compare(a[i], a[j]) == -1 }
 
-// copyBytes returns a copy of b.
-func copyBytes(b []byte) []byte {
-	if b == nil {
-		return nil
-	}
-	buf := make([]byte, len(b))
-	copy(buf, b)
-	return buf
-}
-
 // assert will panic with a given formatted message if the given condition is false.
 func assert(condition bool, msg string, v ...interface{}) {
 	if !condition {
@@ -808,7 +799,7 @@ func assert(condition bool, msg string, v ...interface{}) {
 	}
 }
 
-type byTagKey []*influxql.TagSet
+type byTagKey []*query.TagSet
 
 func (t byTagKey) Len() int           { return len(t) }
 func (t byTagKey) Less(i, j int) bool { return bytes.Compare(t[i].Key, t[j].Key) < 0 }
